"""Basic grid dictionary handling."""

from smmregrid import CdoGrid
from aqua.core.logger import log_configure
from .regridder_util import check_existing_file


class GridDictHandler:
    """
    Class to handle AQUA grid dictionaries.
    It is used to normalize all the different dictionaries and strings 
    that can be used to define a grid.
    It covers many cases (it might seem overdetailed), and can be used also 
    in absence of a cfg_grid_dict dictionary.
    """

    def __init__(self, cfg_grid_dict, default_dimension='2d', loglevel='WARNING'):

        """"
        Initialize the GridDictHandler.
        
        Args:
            cfg_grid_dict (dict): The AQUA grid dictionary from the configuration file.
                                  It can be None or empty dictionary but in this case 
                                  only CDO grid names can be used.
            default_dimension (str): The default dimension to use for the grid path. 
                                     '2d' by default.
            loglevel (str): The logging level.
        """

        self.cfg_grid_dict = cfg_grid_dict
        self.default_dimension = default_dimension
        self.loglevel = loglevel
        self.logger = log_configure(log_level=loglevel, log_name='Regridder')

        # case for no grid dictionary provided
        if not self.cfg_grid_dict:
            self.logger.warning("No cfg_grid_dict dictionary provided, only CDO grid names can be used.")
            return

        # safety checks
        if not isinstance(self.cfg_grid_dict, dict):
            raise ValueError("cfg_grid_dict must be a dictionary.")
        if self.cfg_grid_dict.get('grids') is None:
            raise ValueError("No 'grids' key found in the cfg_grid_dict.")


    def normalize_grid_dict(self, grid_name):
        """
        Normalize the grid name to a grid dictionary, setting grid path as well.

        Args:
            grid_name (str): The grid name (could be a CDO grid).

        Returns:
            dict: The normalized AQUA grid dictionary. 
                  If grid name is none, dictionary with 'path' key as empty dictionary. 
        """

        grid_dict = self._normalize_grid_dict(grid_name)
        grid_dict['path'] = self._normalize_grid_path(grid_dict)
        return grid_dict

    def _normalize_grid_dict(self, grid_name):
        """
        Validate the grid name and return the grid dictionary.
        4 cases handled:
            - None, return an empty dictionary
            - a valid CDO grid name
            - a grid name in the configuration, again a valid CDO grid name
            - a grid dictionary in the configuration

        Args:
            grid_name (str): The grid name.

        Returns:
            dict: The normalized grid dictionary.
        """

        # if empty, return an empty dictionary
        if grid_name is None:
            return {}

        # error if it is not a string or a dictionary
        if not isinstance(grid_name, (str, dict)):
            raise TypeError(f"Grid name '{grid_name}' is not a valid type. str or dict expected.")

        # if a grid name is a valid CDO grid name, return it in the format of a dictionary
        if isinstance(grid_name, str) and CdoGrid(grid_name).grid_kind :
            self.logger.debug("Grid name %s is a valid CDO grid name.", grid_name)
            return {"path": {self.default_dimension: grid_name}}

        # stop if the grid dictionary is not provided
        if self.cfg_grid_dict is None:
            raise ValueError("No cfg_grid_dict dictionary provided, only CDO grid names can be used.")

        # if the grid name is not in the grids
        grid_dict = self.cfg_grid_dict['grids'].get(grid_name)
        if grid_dict is None:
            raise ValueError(f"Grid name '{grid_name}' not found in the configuration.")

        # grid dict is a string: this is the case of a CDO grid name
        if isinstance(grid_dict, str):
            if CdoGrid(grid_dict).grid_kind:
                self.logger.debug("Grid definition %s is a valid CDO grid name.", grid_dict)
                return {"path": {self.default_dimension: grid_dict}}
            raise ValueError(f"Grid name '{grid_dict}' is a string but not a valid CDO grid name.")
        # if the grid dict is a dictionary, return it
        return grid_dict

    def _normalize_grid_path(self, grid_dict):
        """
        Normalize the grid path to a dictionary with the self.default_dimension key.
        3 cases handled: 
            - an empty dictionary, return an empty dictionary
            - a dictionary with a path string, a CDO grid name or a file path
            - a dictionary with a dictionary of path, one for each vertical coordinate

        Args:
            path (str, dict): The grid path.
            data (xarray.Dataset): The dataset to extract grid information from.

        Returns:
            dict: The normalized grid path dictionary. "self.default_dimension" key is mandatory.
        """

        # if empty, return an empty dictionary
        path = grid_dict.get('path')
        if path is None:
            return {}

        # case path is a string: check if it is a valid CDO grid name or a file path
        if isinstance(path, str):
            if CdoGrid(path).grid_kind:
                self.logger.debug("Grid path %s is a valid CDO grid name.", path)
                return {self.default_dimension: path}
            if check_existing_file(path):
                self.logger.debug("Grid path %s is a valid file path.", path)
                return {self.default_dimension: path}
            raise FileNotFoundError(f"Grid file '{path}' does not exist.")

        # case path is a dictionary: check if the values are valid file paths
        # (could extend to CDO names?)
        if isinstance(path, dict):
            for _, value in path.items():
                if not (CdoGrid(value).grid_kind or check_existing_file(value)):
                    raise ValueError(f"Grid path '{value}' is not a valid CDO grid name nor a file path.")
            self.logger.debug("Grid path %s is a valid dictionary of file paths.", path)
            return path

        raise TypeError(f"Grid path '{path}' is not a valid type.")



"""AQUA class for field statitics"""
import xarray as xr
import numpy as np
import regionmask

from smmregrid import GridInspector

from aqua.core.logger import log_configure, log_history
from aqua.core.util import multiply_units

from .area_selection import AreaSelection

# set default options for xarray
xr.set_options(keep_attrs=True)


class FldStat():
    """AQUA class for field statitics"""

    def __init__(self,
                 area: xr.Dataset | xr.DataArray | None = None,
                 horizontal_dims: list[str] | None = None,
                 grid_name: str | None = None,
                 loglevel: str = 'WARNING'):
        """
        Initialize the FldStat.

        Args:
            area (xr.Dataset, xr.DataArray, optional): The area to calculate the statistics for.
            horizontal_dims (list, optional): The horizontal dimensions of the data.
            grid_name (str, optional): The name of the grid, used for logging history.
            loglevel (str, optional): The logging level.
        """
        self.loglevel = loglevel
        self.logger = log_configure(log_level=loglevel, log_name='FldStat')
        self.area = area
        if horizontal_dims is None:
            self.logger.warning("No horizontal dimensions provided, will try to guess from data when provided!")
        self.horizontal_dims = horizontal_dims

        if self.area is None:
            self.logger.warning("No area provided, no weighted area can be provided.")
            return

        # safety checks
        if not isinstance(area, (xr.DataArray, xr.Dataset)):
            raise ValueError("Area must be an xarray DataArray or Dataset.")

        self.grid_name = grid_name

        # Initialize area selection
        self.area_selection = AreaSelection(loglevel=loglevel)

    @property
    def AVAILABLE_FLDSTATS(self):
        """Return available field statistics."""
        return {"custom":   ['integral', 'areasum'],
                "standard": ['mean', 'std', 'max', 'min', 'sum']}

    def fldstat(self, data: xr.DataArray | xr.Dataset,
                stat: str = "mean",
                region: regionmask.Regions | None = None,
                region_sel: str | int | list | None = None,
                mask_kwargs: dict = {},
                lon_limits: list | None = None, lat_limits: list | None = None,
                dims: list | None = None,
                **kwargs):
        """
        Perform a weighted global average.
        If a subset of the data is provided, the average is performed only on the subset.

        Args:
            data (xr.DataArray or xarray.DataDataset):  the input data
            stat (str):  the statistic to compute, only supported is "mean"
            region (regionmask.Regions, optional): A regionmask Regions object defining a class regions.
            region_sel (str, int or list, optional): The region(s) to select by name or number from the region object.
            mask_kwargs (dict, optional): Additional keyword arguments passed to region.mask().
            lon_limits (list, optional):  the longitude limits of the subset
            lat_limits (list, optional):  the latitude limits of the subset
            dims (list, optional):  the dimensions to average over, if not provided, horizontal_dims are used
            **kwargs: additional arguments passed to fldstat

        Kwargs:
            - box_brd (bool,opt): choose if coordinates are comprised or not in area selection.
                                  Default is True

        Returns:
            The value of the averaged field
        """

        if stat not in [s for stats in self.AVAILABLE_FLDSTATS.values() for s in stats]:
            raise ValueError(f"Statistic {stat} not supported by AQUA FldStat(), only "
                             f"{[s for stats in self.AVAILABLE_FLDSTATS.values() for s in stats]} are supported.")

        if not isinstance(data, (xr.DataArray, xr.Dataset)):
            raise ValueError("Data must be an xarray DataArray or Dataset.")

        # if horizontal_dims is not provided, try to guess it
        if self.horizontal_dims is None:
            # please notice GridInspector always return a list of GridType objects
            data_gridtype = GridInspector(data).get_gridtype()
            if len(data_gridtype) > 1:
                raise ValueError("Multiple grid types found in the data, please provide horizontal_dims!")
            self.horizontal_dims = data_gridtype[0].horizontal_dims
            self.logger.debug('Horizontal dimensions guessed from data are %s', self.horizontal_dims)

        # Determine which dims to average over based on mean_type
        if not dims:
            dims = self.horizontal_dims
        else:
            if not isinstance(dims, list):
                raise ValueError("dims must be a list of dimension names.")
            # check if dims are in horizontal_dims
            for dim in dims:
                if dim not in self.horizontal_dims:
                    raise ValueError(f"Dimension {dim} not found in horizontal dimensions: {self.horizontal_dims}")

        # If area is not provided, return the raw mean
        if self.area is None:
            self.logger.warning("No area provided, no area-weighted stat can be provided.")
            # compact call, equivalent of "out = data.mean()"
            if stat in self.AVAILABLE_FLDSTATS["standard"]:
                self.logger.info("Computing unweighted %s on %s dimensions", stat, self.horizontal_dims)
                log_history(data, f"Unweighted {stat} computed on {self.horizontal_dims} dimensions")
                return getattr(data, stat)(dim=self.horizontal_dims)

        # align dimensions naming of area to match data
        self.area = self.align_area_dimensions(data)

        # align coordinates values of area to match data
        self.area = self.align_area_coordinates(data)
    
        if lon_limits is not None or lat_limits is not None or region is not None:
            self.logger.debug("Selecting area for field stat calculation.")
            data = self.area_selection.select_area(data, lon=lon_limits, lat=lat_limits,
                                                   region=region, region_sel=region_sel,
                                                   mask_kwargs=mask_kwargs,
                                                   to_180=False, **kwargs)

        # cleaning coordinates which have "multiple" coordinates in their own definition
        # grid_area = self._clean_spourious_coords(grid_area, name = "area")
        # data = self._clean_spourious_coords(data, name = "data")

        # compact call, equivalent of "out = weighted_data.mean()""
        self.logger.info("Computing area-weighted %s on %s dimensions", stat, dims)

        if stat == 'integral':
            out = self.integrate_over_area(data, self.area, dims)
        elif stat == 'areasum':
            out = self.sum_area(data, self.area, dims)
        elif stat in ['max', 'min']:
            # max/min are not supported by weighted arrays, use unweighted calculation
            out = getattr(data, stat)(dim=dims)
        else:
            weighted_data = data.weighted(weights=self.area.fillna(0))
            out = getattr(weighted_data, stat)(dim=dims)

        if self.grid_name is not None:
            log_history(out, f"From grid '{self.grid_name}'. Computed field stat '{stat}'")

        return out

    def select_area(self, data: xr.Dataset | xr.DataArray,
                    lon: list | None = None, lat: list | None = None,
                    box_brd: bool = True, drop: bool = False,
                    lat_name: str = "lat", lon_name: str = "lon",
                    region: regionmask.Regions | None = None,
                    region_sel: str | int | list | None = None,
                    mask_kwargs: dict = {},
                    default_coords: dict = {"lat_min": -90, "lat_max": 90,
                                            "lon_min": 0, "lon_max": 360},
                    to_180: bool = True) -> xr.Dataset | xr.DataArray:
        """
        Select a specific area from the dataset based on longitude and latitude ranges.
        Wrapper for AreaSelection.select_area method.
        """
        # TODO: The lat_name and lon_name are at the actual stage in the
        # select_area method arguments. However it is possible to foresee
        # that we may want to automatically detect the names
        return self.area_selection.select_area(data, lon=lon, lat=lat,
                                               box_brd=box_brd, drop=drop,
                                               lat_name=lat_name, lon_name=lon_name,
                                               region=region, region_sel=region_sel,
                                               mask_kwargs=mask_kwargs,
                                               default_coords=default_coords, to_180=to_180)

    def integrate_over_area(self, data: xr.Dataset | xr.DataArray,
                            areacell: xr.DataArray,
                            dims: list):
        """
        Compute the integral of the data over the area.

        Args:
            data (xr.DataArray or xr.Dataset): The data, used also for masking.
            areacell (xr.DataArray): The area cells.
            dims (list): Dimensions to sum over.

        Returns:
            xr.DataArray or xr.Dataset: The integral of the data over the area
        """
        area_weighted_data = data * areacell.where(data.notnull())

        # preserve attrs (e.g. AQUA_region) from areacell due to multiplication above, which has priority if keys overlap
        merged_attrs = {**data.attrs, **areacell.attrs}

        if 'units' in data.attrs and 'units' in areacell.attrs:
            merged_attrs['units'] = multiply_units(data.attrs['units'], areacell.attrs['units'])
        else:
            self.logger.warning(f"Data units: {data.attrs.get('units', 'None')}; "
                                f"Area units: {areacell.attrs.get('units', 'None')}, cannot multiply units using Metpy.")

        if 'long_name' in data.attrs:
            merged_attrs['long_name'] = f"Integrated {data.attrs['long_name']}"

        area_weighted_data.attrs.update(merged_attrs)

        area_weighted_integral = area_weighted_data.sum(skipna=True, min_count=1, dim=dims)

        return area_weighted_integral

    def sum_area(self,
                 data: xr.Dataset | xr.DataArray, 
                 areacell: xr.DataArray, 
                 dims: list):
        """
        Compute the sum of area cells where masked data is not null.

        This is useful for computing field such as sea ice extent, by summing
        the area of cells that contain data not null.

        Note: if data is not masked might return incorrect results. If irrelevant regions (e.g., low level or land
        in sea-ice data) are not masked beforehand, their area will be incorrectly included in the sum.

        Args:
            data (xr.DataArray or xr.Dataset): The data (check if pre-masking is needed in the considered variable)
            areacell (xr.DataArray): The area cells.
            dims (list): Dimensions to sum over.

        Returns:
            xr.DataArray or xr.Dataset: The sum of area cells
        """
        summed_area = areacell.where(data).sum(skipna=True, min_count=1, dim=dims)

        if self.grid_name is not None:
            log_history(summed_area, f"Area summed from {self.grid_name} grid")
        return summed_area

    def align_area_dimensions(self, data: xr.Dataset | xr.DataArray):
        """
        Align the area dimensions with the data dimensions.
        If the area and data have different number of horizontal dimensions, try to rename them.

        Args:
            data (xr.DataArray or xr.Dataset): The input data to align with the area.
        """

        # verify that horizontal dimensions area the same in the two datasets.
        # If not, try to rename them. Use gridtype to get the horizontal dimensions
        # TODO: "rgrid" is not a default dimension in smmregrid, it should be added.
        # please notice GridInspector always return a list of GridType objects
        area_gridtype = GridInspector(self.area, extra_dims={"horizontal": ["rgrid"]}).get_gridtype()
        area_horizontal_dims = area_gridtype[0].horizontal_dims
        self.logger.debug("Area horizontal dimensions are %s", area_horizontal_dims)

        if set(area_horizontal_dims) == set(self.horizontal_dims):
            return self.area

        # check if area and data have the same number of horizontal dimensions
        if len(area_horizontal_dims) != len(self.horizontal_dims):
            raise ValueError("Area and data have different number of horizontal dimensions!")

        # check if area and data have the same horizontal dimensions
        self.logger.warning("Area %s and data %s have different horizontal dimensions! Renaming them!",
                            area_horizontal_dims, self.horizontal_dims)
        # create a dictionary for renaming matching dimensions have the same length
        matching_dims = {a: d for a, d in zip(area_horizontal_dims, self.horizontal_dims) if self.area.sizes[a] == data.sizes[d]}
        self.logger.info("Area dimensions has been renamed with %s",  matching_dims)
        return self.area.rename(matching_dims)

    def align_area_coordinates(self, data: xr.Dataset | xr.DataArray, decimals: int = 5):
        """
        Check if the coordinates of the area and data are aligned.
        If they are not aligned, try to flip the coordinates.

        Args:
            data (xr.DataArray or xr.Dataset): The input data to align with the area.
            decimals (int): Number of decimals to use for rounding when aligning coordinates.

        Returns:
            xr.DataArray or xr.Dataset: The area with aligned coordinates.
        """

        # area.coords should be only lon-lat
        for coord in self.area.coords:
            if coord in data.coords and coord != "time":
                area_coord = self.area[coord]
                data_coord = data[coord]

                # verify coordinates has the same sizes
                if area_coord.size != data_coord.size:
                    raise ValueError(f"{coord} has a mismatch in length!")

                # Fast check if coordinates are already aligned
                if np.array_equal(area_coord, data_coord):
                    continue

                # Fast check for reversed coordinates: use slicing
                if np.array_equal(area_coord[::-1], data_coord):
                    self.logger.warning("Reversing coordinate '%s' for alignment.", coord)
                    self.area = self.area.isel({coord: slice(None, None, -1)})
                    continue

                # Try alignment by rounding to specified decimals
                area_rounded = np.round(area_coord, decimals=decimals)
                data_rounded = np.round(data_coord, decimals=decimals)
                if np.array_equal(area_rounded, data_rounded):
                    self.logger.warning("Coordinate '%s' aligned by rounding to %d decimals.", coord, decimals)
                    # assign the rounded coordinates to the area (matching data's rounded values)
                    self.area = self.area.assign_coords({coord: data_coord})
                    continue

                raise ValueError(f"Mismatch in values for coordinate '{coord}' between data and areas.")

        return self.area



"""
This module is used to manage the grid entry in the grid file.
It is used to create the grid entry name, block, and write the grid entry to the grid file.
It also control the filename of the grid file.
"""

import os
import glob
import re
from typing import Optional, Any, Dict
from aqua.core.util import load_yaml, dump_yaml
from aqua.core.configurer import ConfigPath
from aqua.core.logger import log_configure
from aqua.core.lock.safelock import SafeFileLock


class GridEntryManager:
    """
    Class to manage grid entry naming, block creation, and YAML file writing.
    It also controls the filename of the grid file.
    Handles all context for naming and entry logic.
    """

    def __init__(
        self,
        model_name: Optional[str] = None,
        grid_name: Optional[str] = None,
        original_resolution: Optional[str] = None,
        loglevel: str = 'warning'
    ) -> None:
        """
        Initialize the GridEntryManager.

        Args:
            model_name (Optional[str]): The name of the model.
            grid_name (Optional[str]): The name of the grid.
            original_resolution (Optional[str]): The original resolution of the grid.
            loglevel (str): The logging level for the logger.
        """
        # get useful paths
        conf = ConfigPath()
        self.configpath = conf.get_config_dir()
        self.gridpath = os.path.join(self.configpath, 'grids')

        # try to keep model and grid names as lowercase
        self.model_name = model_name.lower() if model_name else None
        self.grid_name = grid_name.lower() if grid_name else None
        self.original_resolution = original_resolution.lower() if original_resolution else None

        # set log level and logger
        self.loglevel = loglevel
        self.logger = log_configure(log_level=loglevel, log_name='GridEntryManager')

    def get_basename(
        self,
        aquagrid: str,
        cdogrid: Optional[str] = None,
        masked: Optional[str] = None,
        vert_coord: Optional[str] = None
    ) -> str:
        """
        Get the basename for the grid type based on the context and aquagrid name.

        Args:
            aquagrid (str): The AQUA grid name.
            cdogrid (Optional[str]): The CDO grid name.
            masked (Optional[str]): Mask type if present.
            vert_coord (Optional[str]): Vertical coordinate if present.
        Returns:
            str: The constructed basename.
        """
        # if masked is not set
        if not masked:
            if cdogrid:
                return aquagrid
            if aquagrid and self.model_name:
                return f"{self.model_name}_{aquagrid}"
            raise ValueError("CDO grid is not detected and model name is not set, please provide model name.")

        # masking
        basename = f"{self.model_name}"
        if self.original_resolution:
            basename += f"_{self.original_resolution}"
        basename += f"_{aquagrid}"
        if vert_coord:
            basename += f"_3d_{vert_coord}"
        else:
            basename += "_2d"
        return basename

    def create_grid_entry_name(
        self,
        aquagrid: str,
        cdogrid: Optional[str] = None,
        masked: Optional[str] = None,
        vert_coord: Optional[str] = None
    ) -> str:
        """
        Create a grid entry name based on the grid type and vertical coordinate.

        Args:
            aquagrid (str): The AQUA grid name.
            cdogrid (Optional[str]): The CDO grid name.
            masked (Optional[str]): Mask type if present.
            vert_coord (Optional[str]): Vertical coordinate if present.
        Returns:
            str: The grid entry name.
        """
        return self.get_basename(aquagrid, cdogrid, masked, vert_coord).replace('_', '-')

    def create_grid_entry_block(
        self,
        path: str,
        horizontal_dims: Optional[str] = None,
        cdo_options: Optional[str] = None,
        remap_method: Optional[str] = None,
        vert_coord: Optional[str] = None
    ) -> Dict[str, Any]:
        """
        Create a grid entry block for the gridtype, with only cdo_options and remap_method.

        Args:
            path (str): Path to the grid file.
            horizontal_dims (Optional[str]): Horizontal dimensions.
            cdo_options (Optional[str]): CDO options string.
            remap_method (Optional[str]): Remap method string.
            vert_coord (Optional[str]): Vertical coordinate if present.
        Returns:
            Dict[str, Any]: The grid entry block.
        """
        grid_block = {
            'path': f"{path}",
        }
        if horizontal_dims:
            grid_block['space_coord'] = horizontal_dims
        if vert_coord:
            grid_block['vert_coord'] = vert_coord
            grid_block['path'] = {vert_coord: f"{path}"}
        if cdo_options:
            grid_block['cdo_options'] = cdo_options
        if remap_method and remap_method != 'con':
            grid_block['remap_method'] = remap_method
        return grid_block

    def get_gridfilename(self, cdogrid: Optional[str], gridkind: str) -> str:
        """
        Get the grid filename based on the grid kind.

        Args:
            cdogrid (Optional[str]): The CDO grid name.
            gridkind (str): The kind of grid.
        Returns:
            str: The grid filename.
        """
        if cdogrid:
            gridfilename = f'{gridkind.lower()}.yaml'
        else:
            gridfilename = f'{self.model_name}-{gridkind.lower()}.yaml'
        return os.path.join(self.gridpath, gridfilename)

    def get_versioned_basepath(self, outdir: str, basename: str, version: Optional[int] = None) -> str:
        """
        Returns the correct basepath (without .nc) for the grid file, handling versioning logic.
        Raises ValueError if versioning rules are violated (e.g., versioned files exist but no version specified).
        Does NOT remove or create any files.

        Args:
            outdir (str): Output directory.
            basename (str): Base name for the file.
            version (Optional[int]): Version number.
        Returns:
            str: The versioned basepath.
        """
        basepath = os.path.join(outdir, basename)
        existing_files = glob.glob(f"{basepath}*.nc")
        if existing_files and version is None:
            pattern = re.compile(r"_v\d+\.nc$")
            check_version = [bool(pattern.search(file)) for file in existing_files]
            if any(check_version):
                raise ValueError(f"Versioned files already exist for {basepath}. Please specify a version")
        if version is not None:
            basepath = f"{basepath}_v{version}"
        return basepath

    def create_grid_entry(self, gridfile: str, grid_entry_name: str,
                          grid_block: Dict[str, Any], rebuild: bool = False) -> None:
        """
        Create or update a grid entry in the grid YAML file.
        Use file locking to prevent race conditions when parallel
        processes try to write to the same grid file simultaneously.

        Args:
            gridfile (str): Path to the grid YAML file.
            grid_entry_name (str): Name of the grid entry.
            grid_block (Dict[str, Any]): The grid entry block.
            rebuild (bool): Whether to overwrite existing entry if present.
        Returns:
            None
        """
        if self.logger:
            self.logger.info("Grid entry name: %s", grid_entry_name)
            self.logger.info("Grid block: %s", grid_block)

        if not os.path.exists(gridfile):
            if self.logger:
                self.logger.info("Grid file %s does not exist, creating it", gridfile)
            final_block = {'grids': {grid_entry_name: grid_block}}
            dump_yaml(gridfile, final_block)
        else:
            lock_path = gridfile + '.lock'
            with SafeFileLock(lock_path, loglevel=self.loglevel):
                if self.logger:
                    self.logger.info("Grid file %s exists, adding the grid entry %s", gridfile, grid_entry_name)
                final_block = load_yaml(gridfile)
                if grid_entry_name in final_block.get('grids', {}) and not rebuild:
                    self.logger.warning("Grid entry %s already exists in %s, skipping", grid_entry_name, gridfile)
                    return
                final_block['grids'][grid_entry_name] = grid_block
                dump_yaml(gridfile, final_block)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
'''
AQUA regridding tool to create low resolution archive.
Make use of aqua.Drop class to perform the regridding.
Functionality can be controlled through CLI options and
a configuration yaml file.
'''

import sys
import argparse
from aqua import Drop
from aqua.core.util import load_yaml, get_arg, to_list
from aqua import __version__ as version


def drop_parser(parser = None):
    """
    Parse command line arguments for the DROP CLI

    Args:
        Optional part to be extended with DROP options
    """

    if parser is None:
        parser = argparse.ArgumentParser(description='AQUA DROP')
    
    parser.add_argument('-c', '--config', type=str,
                        help='yaml configuration file')
    parser.add_argument('-f', '--fix', action="store_true",
                        help='fixer on existing data')
    parser.add_argument('-w', '--workers', type=str,
                        help='number of dask workers')
    parser.add_argument('-d', '--definitive', action="store_true",
                        help='definitive run with files creation')
    parser.add_argument('-o', '--overwrite', action="store_true",
                        help='overwrite existing output')
    parser.add_argument('-l', '--loglevel', type=str,
                        help='log level [default: WARNING]')
    parser.add_argument('--monitoring', action="store_true",
                        help='enable the dask performance monitoring. Will run a single chunk')
    parser.add_argument('--only-catalog', action="store_true",
                        help='To not run DROP but simply create the catalog entries for netcdf and zarr')
    parser.add_argument('--catalog', type=str,
                        help='catalog to be processed. Use with coherence with --model, -exp and --source')
    parser.add_argument('-m', '--model', type=str,
                        help='model to be processed. Use with coherence with --exp and --source')
    parser.add_argument('-e', '--exp', type=str,
                        help='experiment to be processed. Use with coherence with --source and --model')
    parser.add_argument('-s', '--source', type=str,
                        help='source to be processed. Use with coherence with --exp and --var')
    parser.add_argument('-v', '--var', type=str,
                        help='var to be processed. Use with coherence with --source')
    parser.add_argument('--rebuild', action="store_true", help="Rebuild Reader areas and weights")
    parser.add_argument('--realization', type=str,
                        help='realization to be processed. Use with coherence with --model, --exp and --source')
    parser.add_argument('--stat', type=str,
                        help="statistic to be computed. Can be one of ['min', 'max', 'mean', 'std'].")
    parser.add_argument('--frequency', type=str,
                        help="Frequency of the DROP output. Can be anything in the AQUA frequency.")
    parser.add_argument('--resolution', type=str,
                        help="Resolution of the DROP output. Can be anything in the AQUA resolution.")
    parser.add_argument('--startdate', type=str,
                        help="Start date to subset the data. Format YYYY-MM-DD")
    parser.add_argument('--enddate', type=str,
                        help="End date to subset the data. Format YYYY-MM-DD")
    parser.add_argument('--zarr', action="store_true",
                        help='Create zarr')
    parser.add_argument('--verify-zarr', action="store_true",
                        help='Verify the created zarr')  
    #return parser.parse_args(arguments)
    return parser

def drop_execute(args):
    """
    Executing the DROP by parsing the arguments and configuring the machinery
    """

        # to check if GSV is available and return the version
    try:
        import gsv
        print('GSV version is: ' + gsv.__version__)
    except RuntimeError:
        print("GSV not available. FDB5 binary library not present on system.")
    except KeyError:
        print("GSV not available. Environment variables not correctly set.")

    print('AQUA version is: ' + version)

    file = get_arg(args, 'config', 'drop_config.yaml')
    print('Reading configuration yaml file..')

    # basic from configuration
    config = load_yaml(file)

    #safety check
    for item in ['target', 'paths', 'data', 'options']:
        if not item in config:
            raise KeyError(f'Configuration file {file} does not have the "{item}" key, please modify it according to the template')

    # paths
    paths = config['paths']
    outdir = paths['outdir']
    tmpdir = paths['tmpdir']

    # main arguments, ony from config file
    region = config['target'].get('region', None)

    # Command line argumens override config file
    catalog = get_arg(args, 'catalog', config['target'].get('catalog'))
    stat = get_arg(args, 'stat', config['target'].get('stat', 'mean'))
    frequency = get_arg(args, 'frequency', config['target'].get('frequency'))
    resolution = get_arg(args, 'resolution', config['target'].get('resolution'))
    startdate = get_arg(args, 'startdate', config['target'].get('startdate'))
    enddate = get_arg(args, 'enddate', config['target'].get('enddate'))

    loglevel = get_arg(args, 'loglevel', config['options'].get('loglevel', 'WARNING'))
    do_zarr = get_arg(args, 'zarr', config['options'].get('zarr', False))
    verify_zarr = get_arg(args, 'verify_zarr', config['options'].get('verify_zarr', False))

    # Other options, only from command line
    definitive = get_arg(args, 'definitive', False)
    monitoring = get_arg(args, 'monitoring', False)
    overwrite = get_arg(args, 'overwrite', False)
    rebuild = get_arg(args, 'rebuild', False)
    only_catalog = get_arg(args, 'only_catalog', False)
    if only_catalog:
        print('--only-catalog selected, doing a lot of noise but in the end producing only catalog update!')
    fix = get_arg(args, 'fix', True)

    default_workers = get_arg(args, 'workers', 1)

    drop_cli(args=args, config=config, catalog=catalog, resolution=resolution,
            frequency=frequency, fix=fix, enddate=enddate, startdate=startdate,
            outdir=outdir, tmpdir=tmpdir, loglevel=loglevel,
            region=region, stat=stat,
            definitive=definitive, overwrite=overwrite, rebuild=rebuild,
            default_workers=default_workers,
            monitoring=monitoring, do_zarr=do_zarr, verify_zarr=verify_zarr, only_catalog=only_catalog)

def drop_cli(args, config, catalog=None, resolution=None, frequency=None, fix=None,
             startdate=None, enddate=None, outdir=None, tmpdir=None, loglevel=None,
             region=None, stat='mean',
             definitive=False, overwrite=False,
             rebuild=False, monitoring=False,
             default_workers=1, do_zarr=False, verify_zarr=False,
             only_catalog=False):
    """
    Running the default DROP from CLI, looping on all the configuration model/exp/source/var combination
    Optional feature for each source can be defined as `zoom`, `workers` and `realizations`
    Options for dry run and overwriting, as well as monitoring and zarr creation, are available

    Args:
        args: argparse arguments
        config: configuration dictionary
        catalog: catalog to be processed
        resolution: resolution of the DROP output
        frequency: frequency of the DROP output
        fix: fixer option
        outdir: output directory
        tmpdir: temporary directory
        loglevel: log level
        region: region to be processed
        stat: statistic to be computed
        definitive: bool flag to create definitive files
        overwrite: bool flag to overwrite existing files
        rebuild: bool flag to rebuild the areas and weights
        default_workers: default number of workers
        monitoring: bool flag to enable the dask monitoring
        do_zarr: bool flag to create zarr
        verify_zarr: bool flag to verify zarr
        only_catalog: bool flag to only update the catalog
    """

    models = to_list(get_arg(args, 'model', config['data']))
    for model in models:
        exps = to_list(get_arg(args, 'exp', config['data'][model]))
        for exp in exps:

            # if you do require the entire catalog generator
            sources = to_list(get_arg(args, 'source', config['data'][model][exp]))
            for source in sources:
                # get info on potential realizations from the configuration file or from the args of command line
                realizations = get_arg(args, 'realization', config['data'][model][exp][source].get('realizations'))
                loop_realizations = to_list(realizations) if realizations is not None else [1]

                # get info on varlist and workers
                varnames = to_list(get_arg(args, 'var', config['data'][model][exp][source]['vars']))

                # get the number of workers for this specific configuration
                workers = config['data'][model][exp][source].get('workers', default_workers)

                # loop in realizations
                for realization in loop_realizations:

                    # define realization as extra args only if this is found in the configuration file
                    extra_args = {'realization': realization} if realizations else {}
                    print(varnames)
                    for varname in varnames:

                        # get the zoom level - this might need some tuning for extra kwargs
                        zoom = config['data'][model][exp][source].get('zoom', None)
                        if zoom is not None:
                            extra_args = {**extra_args, **{'zoom': zoom}}
                        
                        # disabling rebuild if we are not in the first realization and first varname
                        if varname != varnames[0] or realization != loop_realizations[0]:
                            rebuild = False
                        # init the DROP
                        drop = Drop(catalog=catalog, model=model, exp=exp, source=source,
                                        var=varname, resolution=resolution,
                                        startdate=startdate, enddate=enddate,
                                        frequency=frequency, fix=fix,
                                        outdir=outdir, tmpdir=tmpdir,
                                        nproc=workers, loglevel=loglevel,
                                        region=region, stat=stat,
                                        definitive=definitive, overwrite=overwrite,
                                        rebuild=rebuild,
                                        performance_reporting=monitoring,
                                        exclude_incomplete=True,
                                        **extra_args)


                        
                        if not only_catalog:
                            # check that your DROP output is not already there (it will not work in streaming mode)
                            drop.check_integrity(varname)

                            # retrieve and generate
                            drop.retrieve()
                            drop.drop_generator()

            # create the catalog once the loop is over
            drop.create_catalog_entry()
            if do_zarr:
                drop.create_zarr_entry(verify=verify_zarr)

    print('CLI DROP run completed. Have yourself a tasty pint of beer!')

# if you want to execute the script from terminal without the aqua entry point
if __name__ == '__main__':

    args = drop_parser().parse_args(sys.argv[1:])
    drop_execute(args)
   




"""checksum verification module"""

import os
import sys
import hashlib
import pandas as pd
from aqua.core.util import to_list


def compute_md5(file_path):
    """Compute the MD5 checksum of a file."""
    hash_md5 = hashlib.md5()
    try:
        with open(file_path, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash_md5.update(chunk)
        return hash_md5.hexdigest()
    except FileNotFoundError:
        return None

def generate_checksums(folder, grids, output_file):
    """
    Generate MD5 checksums for all files in a folder.
    Will scan the main folder and the subfolder list and will store the data
    in a output_file
    """

    print(f"Generating datachecker to {output_file}...")
    current_date = pd.Timestamp.now().strftime("%Y-%m-%d %H:%M:%S")
    with open(output_file, "w", encoding='utf8') as f:
        f.write("# MD5 checksum file for AQUA\n")
        f.write(f"# Folder: {folder}\n")
        f.write(f"# Generated by AQUA checksum checker on {current_date} \n\n")

        for grid_path in sorted(grids):
            subdir_path = os.path.join(folder, grid_path)
            if os.path.isdir(subdir_path):
                for root, _, files in os.walk(subdir_path):
                    for file in sorted(files):
                        file_path = os.path.join(root, file)
                        print(f"Computing checksum for {file_path}" )
                        md5_checksum = compute_md5(file_path)
                        if md5_checksum:
                            relative_path = os.path.relpath(file_path, folder)
                            f.write(f"{md5_checksum} {relative_path}\n")
    print(f"Checksum file created at {output_file}")

def verify_checksums(folder, grids, checksum_file):
    """Verify files against MD5 checksums in the checksum file."""

    if isinstance(grids, str):
        if not os.path.exists(os.path.join(folder, grids)):
            raise FileNotFoundError(f'No {grids} directory found in {folder}!')

    try:
        with open(checksum_file, "r", encoding='utf8') as f:
            print(f"Starting verification against {checksum_file}. It will take a while...")
            all_good = True
            for line in f:
                if line.startswith("#") or not line.strip():
                    continue
                md5_checksum, relative_path = line.strip().split(" ", 1)
                relative_dir = os.path.dirname(relative_path)
                if relative_dir in to_list(grids):
                    file_path = os.path.join(folder, relative_path)
                    print(file_path)
                    if not os.path.exists(file_path):
                        print(f"Missing file: {relative_path}!!")
                        all_good = False
                    else:
                        computed_md5 = compute_md5(file_path)
                        if computed_md5 != md5_checksum:
                            print(f"Checksum mismatch for {relative_path}")
                            all_good = False
            if all_good:
                print("All files are verified successfully.")
            else:
                sys.exit("Verification failed.")
    except FileNotFoundError:
        sys.exit(f"Checksum file {checksum_file} not found.")


"""Tests for streaming"""

import pytest
import pandas as pd
from aqua import Reader
from conftest import APPROX_REL, LOGLEVEL

# pytest approximation, to bear with different machines
approx_rel = APPROX_REL
loglevel = LOGLEVEL

# streaming class for tests
@pytest.mark.aqua
class TestAquaStreaming:
    """The streaming testing class"""

    @pytest.fixture(scope="function",
                    params=[{"aggregation": '3S', "startdate": "2020-01-20"},
                            {"aggregation": 'daily', "startdate": "2020-05-01"},
                            {"aggregation": '3D', "startdate": "2020-05-01", "enddate": "2020-05-02"}
                            ])
    def reader_instance_with_args(self, request):
        req = request.param
        req.update({"streaming": True, "model": "IFS", "exp": "test-tco79",
                    "source": "long", "fix": False, "regrid":False})
        return Reader(**req)

    def test_stream(self, reader_instance_with_args):
        """
        Test if the retrieve method returns streamed data with streaming=true
        changing start date and end date
        """

        reader = reader_instance_with_args

        if 'S' in reader.aggregation:
            offset = pd.DateOffset(**{"hours": 3})
        if 'D' in reader.aggregation:
            offset = pd.DateOffset(**{"days": 3})
        if 'daily' in reader.aggregation:
            offset = pd.DateOffset(**{"days": 1})

        step = pd.DateOffset(hours=1)
        
        start_date = pd.to_datetime(reader.startdate)
        if reader.enddate:
            end_date = pd.to_datetime(reader.enddate) + pd.Timedelta(days=1)
        else:
            end_date = start_date + offset

        dates = pd.date_range(start=start_date, end=end_date, freq='h')          
            
        num_hours = (dates[-1] - dates[0]).total_seconds() / 3600
        
        data = reader.retrieve()

         # Test if it has the right size
        assert data['2t'].shape == (num_hours, 9, 18)

        # Test if starting date is ok
        assert data.time.values[0] == pd.to_datetime(start_date)
        
        # Test if end date is ok
        assert data.time.values[-1] == pd.to_datetime(end_date - step)

        # Test if we can go to the next date
        data = reader.retrieve()
        if not reader.enddate:
            assert data.time.values[0] == start_date + offset

        # Test if reset_stream works
        reader.reset_stream()
        data = reader.retrieve()
        assert data.time.values[0] == start_date


"""Class for fixer configuration and loading"""
from aqua.core.logger import log_configure

class FixerConfigure:
    """
    Class to configure the fixer based on convention and fixer_name.
    Args:
        convention (str): The convention name (e.g., 'eccodes'). Default is None
        fixes_dictionary (dict): The fixes dictionary loaded from the fixes file.
        fixer_name (str): The fixer name to load specific fixes. Default is None
        loglevel (str): Log level for logging. Default is 'WARNING'.
    """

    def __init__(self, convention=None, fixes_dictionary=None, fixer_name=None, loglevel='WARNING'):

        self.convention = convention
        self.fixer_name = fixer_name
        self.fixes_dictionary = fixes_dictionary
        self.logger = log_configure(log_level=loglevel, log_name="FixerConfigure")

    def find_fixes(self):
        """
        Get the fixes for the required model, experiment and source.
        A convention dictionary is loaded and merged with the fixer_name.
        The default convention is eccodes-2.39.0.

        If not found, looks for default fixes for the model.
        Then source_fixes are loaded and merged with base_fixes.

        This second block of search is deprecated and will be removed in the future.

        Args:
            The fixer class

        Return:
            The fixer dictionary
        """
        # The convention dictionary is defined by stardard: eccodes and version: 2.39.0
        # At the actual stage 'eccodes' is the default and if not set to None or 'eccodes'
        # an error is raised in the Reader initialization
        convention_dictionary = self._load_convention_dictionary() if self.convention else None

        # Here we load the fixer_name, the specific fix that is merged with the convention dictionary
        # The merge is done only if the base_fixes dictionary has the 'convention' field matching the
        # convention dictionary.
        base_fixes = self._load_fixer_name()
        return self._combine_convention(base_fixes, convention_dictionary)

    def _load_convention_dictionary(self, version='2.39.0'):
        """
        Load the convention dictionary from the fixer file.
        The convention_name block should be called <convention>-<version>.

        Args:
            version: The convention name version. Default is 2.39.0

        Returns:
            The convention dictionary
        """
        convention_dictionary = self.fixes_dictionary.get("convention_name", None)
        if convention_dictionary is None:
            self.logger.error("Convention dictionary not found, will be disabled")
            return None

        convention_name = self.convention + '-' + version

        convention_dictionary = convention_dictionary.get(convention_name, None)
        if convention_dictionary is None:
            self.logger.error("No convention dictionary found for %s", convention_name)
            return None
        else:
            self.logger.info("Convention dictionary: %s", convention_name)

        return convention_dictionary

    def _combine_convention(self, base_fixes: dict, convention_dictionary: dict):
        """
        Combine convention dictionary with the fixes.
        It scan the 'vars' block and merge the convention dictionary with the fixer_name.
        If an item is new in the convention dictionary, it is added to the fixes.
        If the item is already present, non existing keys are added,
        otherwise the existing keys (in the base file) are kept.

        Args:
            base_fixes (dict): The base fixes (coming from the fixer_name)
            convention_dictionary (dict): The convention dictionary

        Returns:
            The final fixes, with the convention dictionary merged
        """
        if base_fixes is None:
            self.logger.info("No fixer_name found, only convention will be applied")
            base_fixes = {}
            base_convention = 'eccodes'
        elif convention_dictionary is None:
            self.logger.info("No convention dictionary found, only fixer_name will be applied")
            return base_fixes
        else:
            base_convention = base_fixes.get('convention', None)
        # We do not crash if the convention is not eccodes, but we log an error and return the base fixes
        convention = convention_dictionary.get('convention', None)
        if convention != 'eccodes':
            self.logger.error("Convention %s not supported, only eccodes is supported", convention)
            return base_fixes

        # We make sure that the convention is the same in the convention dictionary and in the fixer_name
        # Additionally, we check that the version is the same, knowing that for the moment we are not going to
        # document this feature since the version is hardcoded in the code.
        # TODO: version should be a parameter in the fixer_name and in the convention dictionary
        if base_convention is not None:
            if base_convention != convention and convention is not None:
                raise ValueError(f"The convention in the convention dictionary: {base_convention} is different from the fixer_name: {convention}")
            if 'version' in base_fixes and 'version' in convention_dictionary:
                if base_fixes['version'] != convention_dictionary['version']:
                    raise ValueError(f"The version in the convention dictionary: {base_fixes['version']} is different from the fixer_name: {convention_dictionary['version']}")
        else:
            self.logger.info("No convention found in the fixer_name, the convention dictionary will not be used")
            return base_fixes

        # Merge one by one the variables. This is done so that we can be careful at the level of the individual variables.
        # If a field for the specific variable is present in the base_fixes, it has priority.
        if 'vars' in convention_dictionary:
            if 'vars' not in base_fixes:
                base_fixes['vars'] = convention_dictionary['vars']
            else:  # A merge one variable by one is needed
                for var_key in convention_dictionary['vars'].keys():
                    if var_key in base_fixes['vars']:
                        # self.logger.debug("Variable %s already present in the fixes, merging...", var_key)
                        # This requires python >= 3.9
                        base_fixes['vars'][var_key] = convention_dictionary['vars'][var_key] | base_fixes['vars'][var_key]
                        # We need to check that only one between 'source' and 'derived' is present.
                        # If both are present, we give priority to 'derived' and log an info.
                        if 'source' in base_fixes['vars'][var_key] and 'derived' in base_fixes['vars'][var_key]:
                            self.logger.info("Variable %s has both 'source' and 'derived' in the fixes, 'derived' will be used",  # noqa: E501
                                             var_key)
                            base_fixes['vars'][var_key].pop('source')
                    else:
                        # This is a new variable to fix
                        # We need to manipulate the convention_dictionary to set the grib flag
                        # TODO: expand to other formats (cmor tables, etc.)
                        new_var = convention_dictionary['vars'][var_key]
                        if convention == 'eccodes':
                            new_var['grib'] = convention_dictionary['vars'][var_key].get('grib', True)
                        base_fixes['vars'][var_key] = new_var
        else:
            self.logger.warning("No 'vars' block found in the convention dictionary")

        return base_fixes

    def _load_fixer_name(self):
        """
        Load the fixer_name reading from the metadata of the catalog.
        If the fixer_name has a parent, load it and merge it giving priority to the child.
        """
        # if fixer name is found, get it
        if self.fixer_name is not None:
            self.logger.info('Fix names in metadata is %s', self.fixer_name)
            fixes = self.fixes_dictionary["fixer_name"].get(self.fixer_name)
            if fixes is None:
                self.logger.error("The requested fixer_name %s does not exist in fixes files", self.fixer_name)
                return None
            else:
                self.logger.info("Fix names %s found in fixes files", self.fixer_name)

                if 'parent' in fixes:
                    parent_fixes = self.fixes_dictionary["fixer_name"].get(fixes['parent'])
                    if parent_fixes is not None:
                        self.logger.info("Parent fix %s found! Mergin with fixer_name fixes %s!", fixes['parent'],
                                         self.fixer_name)
                        fixes = self._merge_fixes(parent_fixes, fixes)
                    else:
                        self.logger.error("Parent fix %s defined but not available in the fixes file.", fixes['parent'])

            return fixes

        # if not fixes found, return fixes None
        return None

    @staticmethod
    def _merge_fixes(base, specific):
        """
        Small function to merge fixes. Base fixes will be used as a default
        and specific fixes will replace where necessary. Dictionaries will be merged
        for variables, with priority for the specific ones.

        Args:
            base (dict): Base fixes
            specific (dict): Specific fixes

        Return:
            dict with merged fixes
        """
        final = base
        for item in specific.keys():
            if item == 'vars' and item in base:
                final[item] = {**base[item], **specific[item]}
            else:
                final[item] = specific[item]

        return final


"""Test regridding from Reader"""
import pytest
from aqua import Reader, Regridder
from aqua.core.regridder.griddicthandler import GridDictHandler
from conftest import APPROX_REL, LOGLEVEL

approx_rel = APPROX_REL


@pytest.fixture(
    params=[
        ("IFS", "test-tco79", "short", "2t", 0),
        ("IFS", "test-tco79", "short_nn", "2t", 0),
        ("IFS", "test-tco79", "long-nogrid", "mtntrf", 0),
        ("ICON", "test-r2b0", "short", "2t", 0),
        ("ICON", "test-healpix", "short", "2t", 0),
        ("FESOM", "test-pi", "original_2d", "tos", 0.33925926),
        ("NEMO", "test-eORCA1", "long-2d", "tos", 0.3379)
    ]
)
def reader_arguments(request):
    return request.param

# fake dictionary for doing tests
cfg_dict = {
    "grids":{
        "r1000": "r36x18",
        "n128": "n128",
        "tragic": "tragic",
        "afternoon": { "path": { "2d": "r100x50" } },
        "doing": { "path": "noise.nc" },
        "wonderful": { "path": { '2d': "r360x180" } },
        "lovely": { "path": { '2d': "tests" } },
        "amazing": { "path": {
            '2d': "r100x50" ,
            'level': "banana" },
        },
        "romantic": { "path": "n128" },
        "tests": { "path": { "pizza please!" } }
    }
}

@pytest.mark.aqua
class TestRegridder():
    """class for regridding test"""

    def test_grid_dict_handler(self):
        """Test the grid dictionary handler"""

        # generic test
        with pytest.raises(ValueError, match="cfg_grid_dict must be a dictionary."):
            gdh = GridDictHandler("what a wonderful test", loglevel=LOGLEVEL)
        with pytest.raises(ValueError, match="No 'grids' key found in the cfg_grid_dict."):
            gdh = GridDictHandler({"pizza": "margherita"}, loglevel=LOGLEVEL)

        # empty one
        gdh = GridDictHandler(None, loglevel=LOGLEVEL)
        assert gdh.normalize_grid_dict("n256") == {"path": {"2d": "n256"}}
        with pytest.raises(ValueError, match="No cfg_grid_dict dictionary provided, only CDO grid names can be used."):
            gdh.normalize_grid_dict("mywonderfulgrid")

        # empty one
        gdh = GridDictHandler(cfg_grid_dict=cfg_dict, loglevel=LOGLEVEL)
        assert gdh.normalize_grid_dict(None) == {"path": {}}
        # standard
        assert gdh.normalize_grid_dict("r1000") == {"path": {"2d": "r36x18"}}
        # couple of CDO grids
        assert gdh.normalize_grid_dict("n256") == {"path": {"2d": "n256"}}
        assert gdh.normalize_grid_dict("r100x50") == {"path": {"2d": "r100x50"}}

        # test of errors
        with pytest.raises(ValueError, match="Grid name 'ciao' not found in the configuration."):
            gdh.normalize_grid_dict("ciao")
        with pytest.raises(TypeError, match="Grid name '20' is not a valid type."):
            gdh.normalize_grid_dict(20)
        with pytest.raises(ValueError, match="Grid name 'tragic' is a string but not a valid CDO grid name."):
            gdh.normalize_grid_dict("tragic")

        # test on grid path
        assert gdh.normalize_grid_dict("afternoon")['path'] == {'2d': 'r100x50'}
        assert gdh.normalize_grid_dict("lovely")['path'] == {'2d': 'tests'}
        assert gdh.normalize_grid_dict("wonderful")['path'] == {'2d': 'r360x180'}
        assert gdh.normalize_grid_dict("romantic")['path'] == {'2d': 'n128'}
        
        # errors on grid path
        with pytest.raises(FileNotFoundError, match="Grid file 'noise.nc' does not exist."):
            gdh.normalize_grid_dict("doing")
        with pytest.raises(TypeError, match="Grid path '{'pizza please!'}' is not a valid type."):
            gdh.normalize_grid_dict("tests")
        with pytest.raises(ValueError, match="Grid path 'banana' is not a valid CDO grid name nor a file path."):
            gdh.normalize_grid_dict("amazing")
        

    def test_regridder(self):
        """Testing the regridder all in independent way"""

        reader = Reader(catalog='ci', model='IFS', exp='test-tco79', source='long', loglevel='ERROR')
        data = reader.retrieve()
        regridder = Regridder(data=data.isel(time=0), loglevel='debug')

        # Regrid the data
        regridder.weights(tgt_grid_name='r144x72', regrid_method="bil")
        out = regridder.regrid(data)

        assert len(out.lon) == 144
        assert len(out.lat) == 72

        src_area = regridder.areas()
        assert len(src_area.lon) == 18
        assert len(src_area.lat) == 9

        tgt_area = regridder.areas(tgt_grid_name='n32')
        assert len(tgt_area.lon) == 128
        assert len(tgt_area.lat) == 64

    def test_basic_interpolation(self, reader_arguments):
        """
        Test basic interpolation,
        checking output grid dimension and
        fraction of land (i.e. any missing points)
        """
        model, exp, source, variable, ratio = reader_arguments

        reader = Reader(model=model, exp=exp, source=source, regrid="r200",
                        fix=True, loglevel=LOGLEVEL)
        data = reader.retrieve()
        rgd = reader.regrid(data[variable])
        assert len(rgd.lon) == 180
        assert len(rgd.lat) == 90
        assert ratio == pytest.approx((rgd.isnull().sum()/rgd.size).values, rel=approx_rel)  # land fraction

    def test_recompute_weights_fesom2D(self):
        """
        Test interpolation on FESOM, at different grid rebuilding weights,
        checking output grid dimension and fraction of land
        (i.e. any missing points)
        """
        reader = Reader(model='FESOM', exp='test-pi', source='original_2d',
                        regrid='r100', rebuild=True, fix=False, loglevel=LOGLEVEL)
        data = reader.retrieve(var='sst')
        rgd = reader.regrid(data)

        ratio = rgd['sst'].isnull().sum()/rgd['sst'].size  # land fraction

        assert len(rgd.lon) == 360
        assert len(rgd.lat) == 180
        assert len(rgd.time) == 2
        assert 0.33 <= ratio <= 0.36

    def test_recompute_weights_ifs(self):
        """Test the case where no source grid path is specified in the regrid.yaml file
        and areas/weights are reconstructed from the file itself"""
        reader = Reader(model='IFS', exp='test-tco79', source='long',
                        regrid='r100', rebuild=True, loglevel=LOGLEVEL)
        data = reader.retrieve(var='ttr')
        rgd = reader.regrid(data)

        assert len(rgd.lon) == 360
        assert len(rgd.lat) == 180
        assert len(rgd.time) == 4728

    def test_recompute_weights_healpix(self):
        """Test Healpix and areas/weights are reconstructed from the file itself"""
        reader = Reader(model='ICON', exp='test-healpix', source='short',
                        regrid='r100', rebuild=True)
        data = reader.retrieve(var='t')
        rgd = reader.regrid(data)

        assert len(rgd.lon) == 360
        assert len(rgd.lat) == 180
        assert len(rgd.height) == 90
        assert len(rgd.time) == 2

    def test_recompute_weights_fesom3D(self):
        """
        Test interpolation on FESOM, at different grid rebuilding weights,
        checking output grid dimension and fraction of land
        (i.e. any missing points)
        """
        reader = Reader(model='FESOM', exp='test-pi', source='original_3d',
                        regrid='r100', rebuild=True, datamodel=False, fix=False, loglevel=LOGLEVEL)
        data = reader.retrieve(var='temp')
        rgd = reader.regrid(data)

        subdata = rgd.temp.isel(nz1=0)
        ratio1 = subdata.isnull().sum()/subdata.size  # land fraction
        subdata = rgd.temp.isel(nz1=2)
        ratio2 = subdata.isnull().sum()/subdata.size  # land fraction
        assert len(rgd.lon) == 360
        assert len(rgd.lat) == 180
        assert len(rgd.time) == 2
        assert 0.33 <= ratio1 <= 0.36
        assert 0.43 <= ratio2 <= 0.46

    def test_recompute_weights_nemo3D(self):
        """
        Test interpolation on NEMO, at different grid rebuilding weights,
        checking output grid dimension and fraction of land
        (i.e. any missing points)
        """
        reader = Reader(model='NEMO', exp='test-eORCA1', source='short-3d',
                        regrid='r200', rebuild=True, datamodel=False, fix=False, loglevel=LOGLEVEL)
        data = reader.retrieve(var='avg_so')
        rgd = reader.regrid(data)

        subdata = rgd.avg_so.isel(level=0)
        ratio1 = subdata.isnull().sum().values/subdata.size  # land fraction
        subdata = rgd.avg_so.isel(level=6)
        ratio2 = subdata.isnull().sum().values/subdata.size  # land fraction
        assert len(rgd.lon) == 180
        assert len(rgd.lat) == 90
        assert 0.32 <= ratio1 <= 0.36
        assert 0.44 <= ratio2 <= 0.47

    def test_levels_and_regrid(self):
        """
        Test regridding selected levels.
        """
        reader = Reader(model='FESOM', exp='test-pi', source='original_3d',
                        regrid='r100', loglevel=LOGLEVEL)
        data = reader.retrieve()

        layers = [0, 2]
        val = data.aqua.regrid().isel(time=1, nz=2, nz1=layers).wo.aqua.fldmean().values
        assert val == pytest.approx(8.6758228e-08)
        val = data.isel(time=1, nz=2, nz1=layers).aqua.regrid().wo.aqua.fldmean().values
        assert val == pytest.approx(8.6758228e-08)
        val = data.isel(time=1, nz=2, nz1=layers).wo.aqua.regrid().aqua.fldmean().values
        assert val == pytest.approx(8.6758228e-08)
        val = data.isel(time=1, nz=2, nz1=layers).aqua.regrid().thetao.isel(nz1=1).aqua.fldmean().values
        assert val == pytest.approx(274.9045)
        val = data.aqua.regrid().isel(time=1, nz=2, nz1=layers).thetao.isel(nz1=1).aqua.fldmean().values
        assert val == pytest.approx(274.9045)
        val = data.isel(time=1, nz=2, nz1=layers).thetao.aqua.regrid().isel(nz1=1).aqua.fldmean().values
        assert val == pytest.approx(274.9045)

        # test reading specific levels for first vertical coordinate (nz1)
        data = reader.retrieve(level=[2.5, 2275])
        val = data.isel(time=1).aqua.regrid().thetao.isel(nz1=1).aqua.fldmean().values
        assert val == pytest.approx(274.9045)

@pytest.mark.aqua
def test_non_latlon_interpolation():
    """
    Test interpolation to a non regular grid,
    checking appropriate logging message
    """
    reader = Reader(model="IFS", exp="test-tco79", source="short", regrid="F80",
                    fix=True, loglevel=LOGLEVEL, rebuild=True)

    data = reader.retrieve(var='2t')['2t'].isel(time=0).aqua.regrid()

    assert data.shape == (160, 320)
    assert data.values[0, 0] == pytest.approx(246.71156470963325)

@pytest.mark.aqua
def test_regrid_method():
    """Test different regridding method from the grid file"""
    reader = Reader(model="ERA5", exp="era5-hpz3", source="monthly-nn", regrid="r100",
                    fix=True, loglevel=LOGLEVEL, rebuild=True)

    data = reader.retrieve(var='2t')['2t'].isel(time=0).aqua.regrid()

    assert data.shape == (180, 360)
    assert data.values[0, 0] == pytest.approx(242.4824981689453)

    reader = Reader(model="ERA5", exp="era5-hpz3", source="monthly-nn", regrid="r100",
                    fix=True, loglevel=LOGLEVEL, rebuild=True, regrid_method="bil")
    data = reader.retrieve(var='2t')['2t'].isel(time=0).aqua.regrid()

    assert data.shape == (180, 360)
    assert data.values[0, 0] == pytest.approx(252.35510736926696)

# missing test for ICON-Healpix



import numpy as np
import xarray as xr
import builtins
import dask.array as da
from aqua.core.util import convert_data_units
from aqua.core.logger import log_configure

def histogram(data: xr.DataArray, bins = 10, range = None, units = None,
              weighted = True, loglevel='WARNING', dask=True, check=False, density=False):
    """
    Function to calculate a histogram of a DataArray.

    Args:
        data (xarray.Dataset):     The input DataArray. If it is a Dataset, the first variable is used.
        bins (int, optional):      The number of bins for the histogram. Defaults to 10.
        range (tuple, optional):   The lower and upper range of the bins. Defaults to None (in that case it is determined automatically).
        weighted (bool, optional): Use latitudinal weights for the histogram. Defaults to True.
        dask (bool, optional):     If True, uses Dask for parallel computation. Defaults to True.
        units (str, optional):     Convert data to these units. Defaults to None.
        check (bool, optional):    Checks if the sum of counts in the histogram is equal to the size of the data. 
                                   Defaults to False. This forces the histogram to be computed.
        density (bool, optional):  Returns a probability density function,
                                   normalized such that the integral over the range is 1. Defaults to False.
        loglevel (str, optional):  Logging level. Defaults to 'WARNING'.

    Raises:
        TypeError: If the input data is not an xarray DataArray.

    Returns:
        xarray.DataArray: The histogram of the input data.
    """

    if isinstance(data, xr.Dataset):
        data = data[list(data.data_vars.keys())[0]]
    elif not isinstance(data, xr.DataArray):
        raise TypeError('Input data must be an xarray DataArray or Dataset')

    logger = log_configure(log_level=loglevel, log_name='Histogram')

    if units is not None:
        data = convert_data_units(data, var=data.name, units=units, loglevel=loglevel)

    logger.info('Computing histogram with the following parameters: bins={}, range={}'.format(bins, range))

    if weighted:
        logger.debug('Using latitudinal weights')
        if 'lat' not in data.coords:
            raise ValueError("DataArray must have a 'lat' coordinate for weighted histogram.")
        weights = xr.ones_like(data)
        weights = weights * weights.lat
        weights = np.cos(np.radians(weights))
    else:
        weights = None

    if dask and isinstance(data.data, da.Array):
        logger.debug('Using Dask for histogram computation')
        hist, edges = da.histogram(data, weights=weights, bins=bins, range=range, density=density)
    else:
        logger.debug('Using NumPy for histogram computation')
        hist, edges = np.histogram(data, weights=weights, bins=bins, range=range, density=density)
    
    size_of_the_data = data.size

    if check and not density:
        if isinstance(hist, da.Array):
            hist = hist.compute()
        if int(sum(hist)) != size_of_the_data:
            logger.warning('Sum of counts in the histogram is not equal to the size of the data')

    center_of_bin = [ 0.5 * (edges[i] + edges[i+1]) for i in builtins.range(len(edges)-1)]
    width_table = [ edges[i+1] - edges[i] for i in builtins.range(len(edges)-1)]

    counts_per_bin = xr.DataArray(hist, coords=[center_of_bin], dims=["center_of_bin"])
    counts_per_bin = counts_per_bin.assign_coords(width=("center_of_bin", width_table))

    counts_per_bin.attrs = data.attrs

    counts_per_bin.center_of_bin.attrs['units'] = data.units

    counts_per_bin.attrs['size_of_the_data'] = size_of_the_data

    if density:
        if "long_name" in counts_per_bin.attrs:
            counts_per_bin.attrs['long_name'] = 'Pdf of {}'.format(counts_per_bin.attrs['long_name'])
        counts_per_bin.name = 'pdf'
        counts_per_bin.attrs['units'] = 'probability density'
    else:
        if "long_name" in counts_per_bin.attrs:
            counts_per_bin.attrs['long_name'] = 'Histogram of {}'.format(counts_per_bin.attrs['long_name'])
        counts_per_bin.name = 'histogram'
        counts_per_bin.attrs['units'] = 'counts'

    return counts_per_bin


import os
import shutil
import pytest
import xarray as xr
import pandas as pd
from aqua import Drop, Reader
from aqua.core.drop.output_path_builder import OutputPathBuilder
from aqua.core.drop.catalog_entry_builder import CatalogEntryBuilder   
from conftest import LOGLEVEL

DROP_PATH = 'ci/IFS/test-tco79/r1/r100/monthly/mean/global'
DROP_PATH_DAILY = 'ci/IFS/test-tco79/r1/r100/daily/mean/europe'

# pytestmark groups tests that run sequentially on the same worker to avoid conflicts
pytestmark = [
    pytest.mark.aqua,
    pytest.mark.console,
    pytest.mark.xdist_group(name="dask_operations")
]

@pytest.fixture(params=[{"model": "IFS", "exp": "test-tco79", "source": "long", "var": "2t", "outdir": "drop_test"}])
def drop_arguments(request):
    """Provides DROP arguments as a dictionary."""
    return request.param


class TestOutputPathBuilder:
    """Class containing tests for OutputPathBuilder."""

    expected = [
        None,
        'ci/IFS/test-tco79/r1/native/nostat/europe/2t_ci_IFS_test-tco79_r1_native_nostat_europe_202001.nc',
        None,
    ]

    @pytest.mark.parametrize("resolution, frequency, realization, region, stat, expected", [
        ('r100', 'monthly', 'r1', 'global', 'mean', expected[0]),
        (None, None, None, 'europe', None, expected[1]),
        ('r200', 'daily', 'r2', 'global', 'nostat', expected[2]),
    ])
    def test_build_path(self, drop_arguments, resolution, frequency, realization, region, stat, expected):
        """Test building output path."""
        builder = OutputPathBuilder(
            catalog='ci', model=drop_arguments["model"], exp=drop_arguments["exp"],
            resolution=resolution,
            frequency=frequency, realization=realization, stat=stat, region=region)
        path = builder.build_path(
            os.path.join(os.getcwd(), drop_arguments['outdir']),
            var=drop_arguments['var'], year=2020, month=1)

        if not expected:
            DROP_PATH = f'ci/IFS/test-tco79/{realization}/{resolution}/{frequency}/{stat}/{region}'
            expected = os.path.join(os.getcwd(), drop_arguments["outdir"], DROP_PATH,
                                         f"2t_ci_IFS_test-tco79_{realization}_{resolution}_{frequency}_{stat}_{region}_202001.nc")
        else:
            expected = os.path.join(os.getcwd(), drop_arguments["outdir"], expected)

        assert path == expected

class TestCatalogEntryBuilder:
    """Class containing tests for CatalogEntryBuilder."""

    @pytest.mark.parametrize("resolution, frequency, realization, region, stat, source_grid_name, chunks", [
        ('r100', 'monthly', 'r1', 'global', 'mean', 'lon-lat-r100', {'time': 12, 'lat': 180, 'lon': 360}),
        ('r200', 'daily', 'r1', 'global', 'mean', 'lon-lat', {'time': 365, 'lat': 90, 'lon': 180}),
        ('r050s', 'monthly', 'r4', 'europe', 'max', 'lon-lat', {'time': 12, 'lat': 361, 'lon': 720}),
      ])
    def test_create_entry_name(self, drop_arguments, resolution, frequency, realization, region, stat, source_grid_name, chunks):
        """Test creation of entry name."""
        builder = CatalogEntryBuilder(
            catalog='ci', **drop_arguments,
            resolution=resolution, frequency=frequency, realization=realization,
            stat=stat, region=region, loglevel=LOGLEVEL
        )
        entry_name = builder.create_entry_name()
        block = builder.create_entry_details(basedir=drop_arguments["outdir"], source_grid_name=source_grid_name)
        
        if resolution == 'r100' and frequency == 'monthly':
            expected_name = f'lra-{resolution}-{frequency}'
        else:
            expected_name = f'{resolution}-{frequency}'
        
        assert entry_name == expected_name
        assert block['driver'] == 'netcdf'
        assert block['parameters'].keys() == {'realization', 'stat', 'region'}

        builder2 = CatalogEntryBuilder(
            catalog='ci', **drop_arguments,
            resolution=resolution, frequency=frequency, realization='r2',
            stat=stat, region=region, loglevel=LOGLEVEL
        )
        newblock = builder2.create_entry_details(basedir=drop_arguments["outdir"], catblock=block, source_grid_name=source_grid_name)
        assert newblock['args']['urlpath'] == block['args']['urlpath']
        assert newblock['parameters']['realization']['allowed'] == [realization,'r2']
        assert newblock['args']['chunks'] == chunks



class TestDROP:
    """Class containing DROP tests."""

    def test_definitive_false(self, drop_arguments, tmp_path):
        """Test DROP with definitive=False."""
        test = Drop(
            catalog='ci', **drop_arguments, tmpdir=str(tmp_path),
            resolution='r100', frequency='monthly', loglevel=LOGLEVEL
        )

        test.retrieve()
        test.drop_generator()
        assert os.path.isdir(os.path.join(os.getcwd(), drop_arguments["outdir"], DROP_PATH))
        shutil.rmtree(os.path.join(drop_arguments["outdir"]))

    @pytest.mark.parametrize("nworkers", [1, 2])
    def test_definitive_true(self, drop_arguments, tmp_path, nworkers):
        test = Drop(
            catalog='ci', **drop_arguments, tmpdir=str(tmp_path),
            nproc=nworkers, resolution='r100', frequency='monthly',
            definitive=True, loglevel=LOGLEVEL
        )

        test.retrieve()
        test.data = test.data.sel(time="2020-01")
        test.drop_generator()

        file_path = os.path.join(os.getcwd(), drop_arguments["outdir"], DROP_PATH, "2t_ci_IFS_test-tco79_r1_r100_monthly_mean_global_202001.nc")
        test.check_integrity(varname=drop_arguments["var"])
        assert os.path.isfile(file_path)

        file = xr.open_dataset(file_path)
        assert len(file.time) == 1
        assert pytest.approx(file['2t'][0, 1, 1].item()) == 248.0704
        shutil.rmtree(os.path.join(drop_arguments["outdir"]))

    def test_regional_subset(self, drop_arguments, tmp_path):
        """Test DROP with regional subset."""
        region = {'name': 'europe', 'lon': [-10, 30], 'lat': [35, 70]}

        test = Drop(
            catalog='ci', **drop_arguments, tmpdir=str(tmp_path),
            resolution='r100', frequency='daily', definitive=True,
            loglevel=LOGLEVEL, region=region
        )

        test.retrieve()
        test.data = test.data.sel(time="2020-01-20")
        test.drop_generator()

        file_path = os.path.join(os.getcwd(), drop_arguments["outdir"], DROP_PATH_DAILY, "2t_ci_IFS_test-tco79_r1_r100_daily_mean_europe_202001.nc")
        assert os.path.isfile(file_path), "File not found: {}".format(file_path)

        xfield = xr.open_dataset(file_path).where(lambda x: x.notnull(), drop=True)
        assert xfield.lat.min() > 35
        assert xfield.lat.max() < 70
        shutil.rmtree(os.path.join(drop_arguments["outdir"]))

    def test_zarr_entry(self, drop_arguments, tmp_path):
        """Test DROP with Zarr archive creation."""
        test = Drop(
            catalog='ci', **drop_arguments, tmpdir=str(tmp_path),
            resolution='r100', frequency='monthly', nproc=1,
            loglevel=LOGLEVEL, definitive=True,
            startdate="2020-01-01", enddate="2020-05-31"
        )

        test.retrieve()
        test.drop_generator()
        test.create_catalog_entry()
        test.create_zarr_entry()

        reader1 = Reader(model=drop_arguments["model"], exp=drop_arguments["exp"], source='lra-r100-monthly')
        reader2 = Reader(model=drop_arguments["model"], exp=drop_arguments["exp"], source='r100-monthly-zarr')

        data1 = reader1.retrieve()
        data2 = reader2.retrieve()
        assert data1.equals(data2)
        shutil.rmtree(os.path.join(drop_arguments["outdir"]))

    def test_dask_overwrite(self, drop_arguments, tmp_path):
        """Test DROP with overwrite=True and Dask initialization."""
        test = Drop(
            catalog='ci', **drop_arguments, tmpdir=str(tmp_path),
            resolution='r100', frequency='monthly', nproc=4,
            loglevel=LOGLEVEL, definitive=True, overwrite=True
        )

        test.retrieve()
        test.drop_generator()
        assert os.path.isdir(os.path.join(os.getcwd(), drop_arguments["outdir"], DROP_PATH))
        shutil.rmtree(os.path.join(drop_arguments["outdir"]))

    def test_exclude_incomplete(self, drop_arguments, tmp_path):
        """Test DROP's exclude_incomplete option."""
        test = Drop(
            catalog='ci', **drop_arguments, tmpdir=str(tmp_path),
            resolution='r100', frequency='monthly', definitive=True,
            loglevel=LOGLEVEL, exclude_incomplete=True
        )

        test.retrieve()
        test.drop_generator()

        missing_file = os.path.join(os.getcwd(), drop_arguments["outdir"], DROP_PATH, "2t_ci_IFS_test-tco79_r1_r100_monthly_mean_global_202008.nc")
        existing_file = os.path.join(os.getcwd(), drop_arguments["outdir"], DROP_PATH, "2t_ci_IFS_test-tco79_r1_r100_monthly_mean_global_202002.nc")

        assert not os.path.exists(missing_file)
        assert os.path.exists(existing_file)

        file = xr.open_dataset(existing_file)
        assert len(file.time) == 1
        test.check_integrity(varname=drop_arguments["var"])
        shutil.rmtree(os.path.join(drop_arguments["outdir"]))

    def test_concat_var_year(self, drop_arguments, tmp_path):
        """Test concatenation of monthly files into a single yearly file."""
        resolution = 'r100'
        frequency = 'monthly'
        year = 2022

        test = Drop(
            catalog='ci', **drop_arguments, tmpdir=str(tmp_path),
            resolution=resolution, frequency=frequency, loglevel=LOGLEVEL
        )

        for month in range(1, 13):
            mm = f'{month:02d}'
            filename = test.get_filename(drop_arguments["var"], year, month=mm)
            timeobj = pd.Timestamp(f'{year}-{mm}-01')
            ds = xr.Dataset({drop_arguments["var"]: xr.DataArray([0], dims=['time'], coords={'time': [timeobj]})})
            ds.to_netcdf(filename)

        test._concat_var_year(drop_arguments["var"], year)
        outfile = test.get_filename(drop_arguments["var"], year)

        assert os.path.exists(outfile)
        shutil.rmtree(os.path.join(drop_arguments["outdir"]))

    def test_concat_var_year_cdo(self, drop_arguments, tmp_path):
        """Test concatenation of monthly files into a single yearly file using cdo."""
        resolution = 'r100'
        frequency = 'monthly'
        year = 2022

        test = Drop(
            catalog='ci', **drop_arguments, tmpdir=str(tmp_path), compact="cdo",
            resolution=resolution, frequency=frequency, loglevel=LOGLEVEL
        )

        for month in range(1, 13):
            mm = f'{month:02d}'
            filename = test.get_filename(drop_arguments["var"], year, month=mm)
            timeobj = pd.Timestamp(f'{year}-{mm}-01')
            ds = xr.Dataset({drop_arguments["var"]: xr.DataArray([0], dims=['time'], coords={'time': [timeobj]})})
            ds.to_netcdf(filename)

        test._concat_var_year(drop_arguments["var"], year)
        outfile = test.get_filename(drop_arguments["var"], year)

        assert os.path.exists(outfile)
        shutil.rmtree(os.path.join(drop_arguments["outdir"]))


"""Simple catalog utility"""
from aqua.core.configurer import ConfigPath


def show_catalog_content(catalog=None, model=None, exp=None, source=None, configdir=None, catalog_name=None,
                         loglevel='WARNING', verbose=True, show_descriptions=False):
    """
    Display the catalog content structure (model/exp/source) without requiring
    manual ConfigPath instantiation.

    This is a convenience wrapper around ConfigPath.show_catalog_content() that
    handles the ConfigPath initialization internally.

    Args:
        catalog (str | list | None): Specific catalog(s) to scan. If None, loops over all available catalogs.
        model (str | None): Optional model filter. If provided, only shows entries for this model.
        exp (str | None): Optional experiment filter. If provided, only shows entries for this exp.
        source (str | None): Optional source filter. If provided, only shows entries for this source.
        configdir (str, optional): The directory containing the configuration files. If not provided, ConfigPath will determine it automatically.
        catalog_name (str, optional): Override the catalog name. If not provided, uses the default catalog.
        loglevel (str, optional): Logging level. Defaults to 'WARNING'.
        verbose (bool): If True, prints the formatted catalog structure. Defaults to True.
        show_descriptions (bool): If True, also print per-source descriptions.

    Returns:
        dict: Dictionary with catalog names as keys and nested dict structure
              as values.
    """
    config = ConfigPath(configdir=configdir, catalog=catalog_name, loglevel=loglevel)
    return config.show_catalog_content(catalog=catalog, model=model, exp=exp, source=source, verbose=verbose,
                                       show_descriptions=show_descriptions)



import numpy as np
import xarray as xr
import pytest
import dask.array as da
from aqua import histogram

@pytest.fixture
def sample_data():
    """
    Fixture to create a sample xarray.DataArray for testing.
    """
    lats = np.arange(0, 10, 1)
    lons = np.arange(0, 10, 1)
    data = xr.DataArray(np.random.rand(10, 10), coords=[('lat', lats), ('lon', lons)], name='test_data')
    data.attrs['units'] = 'm'
    data.attrs['long_name'] = 'Test Data'
    return data

@pytest.fixture
def sample_dask_data(sample_data):
    """
    Fixture to create a sample xarray.DataArray with a dask array for testing.
    """
    data = sample_data.copy()
    data.data = da.from_array(data.data, chunks=(5, 5))
    return data

@pytest.fixture
def sample_dataset(sample_data):
    """
    Fixture to create a sample xarray.Dataset for testing.
    """
    ds = sample_data.to_dataset()
    ds['second_var'] = xr.DataArray(np.random.rand(10, 10), coords=sample_data.coords)
    return ds

@pytest.mark.aqua
def test_histogram_basic(sample_data):
    """
    Test the basic functionality of the histogram function.
    """
    hist = histogram(sample_data, bins=5, check=True, range=(0, 1), weighted=False)
    assert isinstance(hist, xr.DataArray)
    assert hist.name == 'histogram'
    assert 'center_of_bin' in hist.coords
    assert int(hist.sum()) == sample_data.size

@pytest.mark.aqua
def test_histogram_density(sample_data):
    """
    Test the histogram function with density=True.
    """
    hist = histogram(sample_data, bins=5, density=True, range=(0, 1), weighted=False)
    assert isinstance(hist, xr.DataArray)
    assert hist.name == 'pdf'
    # The integral of the PDF should be close to 1
    assert np.isclose(hist.sum() * (hist.center_of_bin[1] - hist.center_of_bin[0]), 1, atol=1e-5)

@pytest.mark.aqua
def test_histogram_weighted(sample_data):
    """
    Test the histogram function with weighted=True.
    """
    hist_unweighted = histogram(sample_data, bins=5, weighted=False, check=True, range=(0, 1))
    hist_weighted = histogram(sample_data, bins=5, weighted=True, check=True, range=(0, 1))
    assert not np.allclose(hist_unweighted.values, hist_weighted.values)

@pytest.mark.aqua
def test_histogram_units(sample_data):
    """
    Test the histogram function with units conversion.
    """
    hist = histogram(sample_data, units='cm', bins=5, range=(0, 100))
    assert hist.center_of_bin.attrs['units'] == 'cm'

@pytest.mark.aqua
def test_histogram_type_error():
    """
    Test that a TypeError is raised for invalid input data.
    """
    with pytest.raises(TypeError):
        histogram(np.random.rand(10, 10))

@pytest.mark.aqua
def test_histogram_dask(sample_dask_data):
    """
    Test the histogram function with a dask array.
    """
    hist = histogram(sample_dask_data, bins=5, check=True, range=(0, 1), weighted=False)
    assert isinstance(hist, xr.DataArray)
    assert hist.name == 'histogram'
    assert 'center_of_bin' in hist.coords
    # The result of histogram with dask is a dask array, so we need to compute it
    assert int(hist.sum().compute()) == sample_dask_data.size

@pytest.mark.aqua
def test_histogram_dataset(sample_dataset):
    """
    Test the histogram function with an xarray.Dataset.
    """
    hist = histogram(sample_dataset, bins=5, check=True, range=(0, 1), weighted=False)
    assert isinstance(hist, xr.DataArray)
    assert hist.name == 'histogram'
    assert 'center_of_bin' in hist.coords
    assert int(hist.sum()) == sample_dataset['test_data'].size

@pytest.mark.aqua
def test_histogram_weighted_no_lat():
    """
    Test that a ValueError is raised when weighted=True and 'lat' coordinate is missing.
    """
    data = xr.DataArray(np.random.rand(10, 10), dims=['x', 'y'], name='test_data')
    data.attrs['units'] = 'm'
    data.attrs['long_name'] = 'Test Data'
    with pytest.raises(ValueError):
        histogram(data, weighted=True)



"""Catalog configuration helpers for AQUA."""
import os
#import platform
import intake

from aqua.core.util.util import to_list
from aqua.core.util.yaml import load_yaml
from aqua.core.logger import log_configure
from .locator import ConfigLocator



class ConfigPath():
    """
    A class to manage the configuration path and directory robustly, including
    handling and browsing across multiple catalogs.
    """

    def __init__(self, configdir=None, filename='config-aqua.yaml',
                 catalog=None, loglevel='warning', locator=None):
        """
        Initialize the ConfigPath instance.
        Args:
            configdir (str | None): The directory where the configuration file is located.
                                        If None, it is determined by the `get_config_dir` method.
            filename (str): The name of the configuration file. Defaults to 'config-aqua.yaml'.
            catalog (str | list | None): Specific catalog(s) to use. If None,
                                        all available catalogs are considered.
            loglevel (str): The logging level. Defaults to 'warning'.
            locator (ConfigLocator | None): An optional ConfigLocator instance.
        """

        # set up logger
        self.logger = log_configure(log_level=loglevel, log_name='ConfigPath')

        # get the configuration directory and its file
        self.filename = filename
        if locator is None:
            locator = ConfigLocator(filename=filename, configdir=configdir, logger=self.logger)
        self.locator = locator
        self.configdir = self.locator.configdir
        self.config_file = self.locator.config_file
        self.logger.debug('Configuration file found in %s', self.config_file)
        self.config_dict = load_yaml(self.config_file)

        # if no catalog are provided, get all available
        if catalog is None:
            catalog = self.get_catalog()
        self.catalog_available = to_list(catalog)
        self.logger.debug('Available catalogs are %s', self.catalog_available)

        # set the catalog as the first available and get all configurations
        if not self.catalog_available:
            self.logger.warning('No available catalogs found')
            self.catalog = None
            self.base_available = None
            self.catalog_file = None
            self.machine_file = None
        else:
            self.catalog = self.catalog_available[0]
            self.base_available = self.get_base()
            self.logger.debug('Default catalog will be %s', self.catalog)
            self.catalog_file, self.machine_file = self.get_catalog_filenames(self.catalog)

        # get also info on machine on init
        self.machine = self.get_machine()

    def get_config_dir(self):
        """
        Return the path to the configuration directory.

        Notes:
            This method delegates to `ConfigLocator` and is kept for backward
            compatibility.
        """
        return self.locator.configdir

    def get_catalog(self):
        """
        Extract the name of the catalog from the configuration file

        Returns:
            list[str] | None: the catalog names from the main config file or
            None when the `catalog` entry is present but empty.
        """
        if os.path.exists(self.config_file):
            base = load_yaml(self.config_file)
            if 'catalog' not in base:
                raise KeyError(f'Cannot find catalog information in {self.config_file}')

            # particular case of an empty list
            if not base['catalog']:
                return None

            self.logger.debug('Catalog found in %s file are %s', self.config_file, base['catalog'])
            return base['catalog']

        raise FileNotFoundError(f'Cannot find the basic configuration file {self.config_file}!')

    def browse_catalogs(self, model: str, exp: str, source: str):
        """
        Given a triplet of model-exp-source, browse all catalog installed catalogs

        Returns
            a list of catalogs where the triplet is found
            a dictionary with information on wrong triplet
        """
        success = []
        fail = {}

        if self.catalog_available is None:
            return success, fail

        if not all(v is not None for v in [model, exp, source]):
            raise KeyError('Need to defined the triplet model, exp and source')

        for catalog in self.catalog_available:
            self.logger.debug('Browsing catalog %s ...', catalog)
            catalog_file, _ = self.get_catalog_filenames(catalog)
            cat = intake.open_catalog(catalog_file)
            check, level, avail = self.scan_catalog(cat, model=model, exp=exp, source=source)
            if check:
                self.logger.info('%s_%s_%s triplet found in in %s!', model, exp, source, catalog)
                success.append(catalog)
            else:
                fail[catalog] = (f'In catalog {catalog} when looking for {model}_{exp}_{source} '
                                 f'triplet I could not find the {level}. Available alternatives are {avail}')
        return success, fail

    def deliver_intake_catalog(self, model, exp, source, catalog=None):
        """
        Given a triplet of model-exp-source (and possibly a specific catalog), browse the catalog
        and check if the triplet can be found

        Returns:
            intake.catalog.Catalog: The intake catalog
            str: The path to the catalog file
            str: The path to the machine file
        """
        matched, failed = self.browse_catalogs(model=model, exp=exp, source=source)
        if not matched:
            for _, value in failed.items():
                self.logger.error(value)
            raise KeyError('Cannot find the triplet in any catalog. Check logger error for hints on possible typos')

        if catalog is not None:
            self.catalog = catalog
        else:
            if len(matched)>1:
                self.logger.warning('Multiple triplets found in %s, setting %s as the default', matched, matched[0])
            self.catalog = matched[0]

        self.logger.debug('Final catalog to be used is %s', self.catalog)
        self.catalog_file, self.machine_file = self.get_catalog_filenames(self.catalog)
        return intake.open_catalog(self.catalog_file), self.catalog_file, self.machine_file

    def get_machine_info(self):
        """
        Extract the information related to the machine from the catalog-dependent machine file

        Returns:
            machine_paths (dict): the machine_paths filesystem locations
            intake_vars (dict): the intake catalog variables
        """
        # loading the grid defintion file
        machine_file = load_yaml(self.machine_file)
        machine_paths = {}

        # get information on paths
        if self.machine in machine_file:
            machine_paths = machine_file[self.machine]
        else:
            if 'default' in machine_file:
                machine_paths = machine_file['default']

        # The main config file has priority
        if 'paths' in self.config_dict:
            for path in ['areas', 'weights', 'grids']:
                if path in self.config_dict['paths']:
                    if 'paths' not in machine_paths:
                        machine_paths['paths'] = {}
                    machine_paths['paths'][path] = self.config_dict['paths'][path]
        else:
            self.logger.debug('No paths found in the main configuration file %s', self.base_available)
        if machine_paths == {}:
            self.logger.error('Cannot find machine paths for %s, regridding and areas feature will not work', self.machine)

        # extract potential intake variables
        intake_vars = machine_paths.get('intake', {})
        return machine_paths, intake_vars

    def get_base(self):
        """
        Get all the possible base configurations available

        Returns:
            dict[str, dict]: map of catalog name to rendered configuration.
        """
        if os.path.exists(self.config_file):
            base = {}
            for catalog in self.catalog_available:
                definitions = {'catalog': catalog, 'configdir': self.configdir}
                base[catalog] = load_yaml(infile=self.config_file, definitions=definitions, jinja=True)
            return base
        raise FileNotFoundError(f'Cannot find the basic configuration file {self.config_file}!')

    def get_machine(self):
        """
        Extract the name of the machine from the configuration file

        Returns:
            str | None: resolved machine name from the configuration file, or None when detection fails.
        """
        if not os.path.exists(self.config_file):
            raise FileNotFoundError(f'Cannot find the basic configuration file {self.config_file}!')

        base = load_yaml(self.config_file)
        # if we do not know the machine we assume is "unknown"
        machine = 'unknown'
        # if the configuration file has a machine entry, use it
        if 'machine' in base:
            self.logger.debug('Machine found in configuration file, set to %s', machine)
            return base['machine']
        
        # warning for unknown machine
        self.logger.warning('No machine entry found in configuration file, set to %s', machine)
        return machine
    
        # if the entry is auto, or the machine unknown, try autodetection
        # if self.machine in ['auto', 'unknown']:
        #     self.logger.debug('Machine is %s, trying to self detect', self.machine)
        #     self.machine = self._auto_detect_machine()

    # def _auto_detect_machine(self):
    #     """Tentative method to identify the machine from the hostname"""

    #     platform_name = platform.node()

    #     if os.getenv('GITHUB_ACTIONS'):
    #         self.logger.debug('GitHub machine identified!')
    #         return 'github'

    #     platform_dict = {
    #         'uan': 'lumi',
    #         'levante': 'levante',
    #     }

    #     # Search for the dictionary key in the key_string
    #     for key, value in platform_dict.items():
    #         if key in platform_name:
    #             self.logger.debug('%s machine identified!', value)
    #             return value

    #     self.logger.debug('No machine identified, still unknown and set to None!')
    #     return None

    def get_catalog_filenames(self, catalog=None):
        """
        Extract the catalog and machine file paths for the selected catalog.

        Args:
            catalog (str | None): override catalog to inspect; defaults to the
                current `self.catalog`.

        Returns:
            catalog_file (str): the path to the catalog file
            machine_file (str): the path to the machine file
        """
        if self.catalog is None:
            raise KeyError('No AQUA catalog is installed. Please run "aqua add CATALOG_NAME"')

        if catalog is None:
            catalog = self.catalog

        catalog_file = self.base_available[catalog]['reader']['catalog']
        self.logger.debug('Catalog file is %s', catalog_file)
        if not os.path.exists(catalog_file):
            raise FileNotFoundError(f'Cannot find catalog file in {catalog_file}. Did you install it with "aqua add {catalog}"?')

        machine_file = self.base_available[catalog]['reader']['machine']
        self.logger.debug('Machine file is %s', machine_file)
        if not os.path.exists(machine_file):
            raise FileNotFoundError(f'Cannot find machine file for {catalog} in {machine_file}')

        return catalog_file, machine_file

    def get_reader_filenames(self, catalog=None):
        """
        Extract the filenames for the reader for catalog, regrid and fixer

        Returns:
            Three strings for the path of the fixer, regrid and config files
        """
        if catalog is None:
            catalog = self.catalog

        fixer_folder = self.base_available[catalog]['reader']['fixer']
        if not os.path.exists(fixer_folder):
            raise FileNotFoundError(f'Cannot find the fixer folder in {fixer_folder}')
        grids_folder = self.base_available[catalog]['reader']['regrid']
        if not os.path.exists(grids_folder):
            raise FileNotFoundError(f'Cannot find the regrid folder in {grids_folder}')

        return fixer_folder, grids_folder


    def scan_catalog(self, cat, model=None, exp=None, source=None):
        """
        Check if the model, experiment and source are in the catalog.

        Returns:
            status (bool): True if the triplet is found
            level (str): The level at which the triplet is failing
            info (str): The available catalog entries at the level of the triplet
        """
        status = False
        avail = None
        level = None

        if model in cat:
            if exp in cat[model]:
                if source in cat[model][exp]:
                    status = True
                else:
                    level = 'source'
                    avail = list(cat[model][exp].keys())
            else:
                level = 'exp'
                avail = list(cat[model].keys())
        else:
            level = 'model'
            avail = list(cat.keys())

        return status, level, avail


    def show_catalog_content(self, catalog=None, model=None, exp=None, source=None, verbose=True,
                             show_descriptions=False):
        """
        Scan catalog(s) by reading YAML files directly and display the model/exp/source structure.
        Uses intake to handle path resolution automatically.

        Args:
            catalog (str | list | None): Specific catalog(s) to scan. If None, loops over all available catalogs.
            model (str | None): Optional model filter.
            exp (str | None): Optional experiment filter.
            source (str | None): Optional source filter.
            verbose (bool): If True, prints the formatted catalog structure. Defaults to True.
            show_descriptions (bool): If True, also print per-source descriptions.

        Returns:
            dict: Dictionary with catalog names as keys and nested dict structure as values.
        """
        self.logger = log_configure(log_level='info', log_name='ShowCatalog')

        results = {}
        catalogs_to_scan = to_list(catalog) if catalog else self.catalog_available

        if not catalogs_to_scan:
            self.logger.warning('No catalogs available to scan')
            return results
        
        self.logger.debug("Catalogs to show: %s", catalogs_to_scan)

        for cat_name in catalogs_to_scan:

            try:
                catalog_file, _ = self.get_catalog_filenames(catalog=cat_name)
                cat = intake.open_catalog(catalog_file)
            except (KeyError, FileNotFoundError, Exception) as exc:
                self.logger.warning('Cannot open/scan catalog %s: %s', cat_name, exc)
                continue

            structure = {}
            descriptions = {}  # model -> exp -> {source: description}
            
            models = [model] if model else list(cat.keys())

            for model_name in models:
                if model_name not in cat:
                    self.logger.warning('Model %s not found in catalog %s. Available: %s', 
                                        model_name, cat_name, list(cat.keys()))
                    continue

                model_cat = cat[model_name]
                experiments = [exp] if exp else list(model_cat.keys())

                for exp_name in experiments:
                    if exp_name not in model_cat:
                        self.logger.warning('Experiment %s not found in model %s. Available: %s', 
                                            exp_name, model_name, list(model_cat.keys()))
                        continue

                    exp_cat = model_cat[exp_name]
                    sources = list(exp_cat.keys())

                    if source:
                        sources = [s for s in sources if s == source]

                    if not sources:
                        continue

                    if model_name not in structure:
                        structure[model_name] = {}
                    structure[model_name][exp_name] = sources

                    # Pre-fetch descriptions if needed (once per experiment)
                    if show_descriptions:
                        if model_name not in descriptions:
                            descriptions[model_name] = {}
                        descriptions[model_name][exp_name] = self._extract_source_descriptions(exp_cat, sources)

            if not structure:
                continue

            results[cat_name] = structure

            if verbose:
                print(self.format_catalog_structure(structure, cat_name, descriptions if show_descriptions else None))

        return results


    @staticmethod
    def _extract_source_descriptions(exp_cat, sources):
        """Safely extract descriptions for a list of sources from an experiment catalog."""
        try:
            walk_dict = exp_cat.walk()
        except Exception:  # noqa: BLE001
            return {}
            
        descs = {}
        for source in sources:
            try:
                # Access internal _description as in original implementation
                desc = walk_dict[source]._description  # pylint: disable=W0212
                if desc:
                    descs[source] = desc
            except Exception:  # noqa: BLE001
                pass
        return descs


    @staticmethod
    def format_catalog_structure(structure, catalog_name, descriptions=None):
        """
        Format catalog structure as a nicely aligned tree.
        
        Args:
            structure: Dictionary with model/exp/source structure
            catalog_name: Name of the catalog
            descriptions: Optional nested dict [model][exp][source] -> description.
                          If provided, prints one source per line with description.
                          If None, prints compact 3-column view.
        """
        lines = [f"\n{'='*80}", f"📁 Catalog: {catalog_name}", f"{'='*80}"]

        for model_name, experiments in sorted(structure.items()):
            lines.append(f"\n   Model: {model_name}")

            for exp_name, sources in sorted(experiments.items()):
                lines.append(f"     └─ Experiment: {exp_name}")

                if not sources:
                    continue
                    
                sorted_sources = sorted(sources)

                if descriptions:
                    # Detailed view: One source per line with description
                    exp_descs = descriptions.get(model_name, {}).get(exp_name, {})
                    for source_key in sorted_sources:
                        desc = exp_descs.get(source_key, "")
                        lines.append(f"        - {source_key:<25} {desc}")
                else:
                    # Compact view: Group sources in rows of 3
                    for i in range(0, len(sorted_sources), 3):
                        source_group = sorted_sources[i:i+3]
                        formatted_sources = "  ".join(f"{s:<25}" for s in source_group)
                        lines.append(f"        ├─ {formatted_sources}")

        lines.append(f"{'='*80}\n")
        return "\n".join(lines)


from typing import Optional, Tuple
import matplotlib.pyplot as plt
import xarray as xr
from aqua.core.util import to_list, unit_to_latex
from aqua.core.logger import log_configure
from .styles import ConfigStyle


def plot_vertical_lines(data: xr.DataArray | list[xr.DataArray],
                        ref_data: xr.DataArray = None,
                        lev_name: str = "plev",
                        lev_min: Optional[float] = None, lev_max: Optional[float] = None,
                        labels: Optional[list[str]] = None,
                        ref_label: Optional[str] = None,
                        label_size: Optional[int] = 14,
                        axis_label_size: Optional[int] = 14,
                        title: Optional[str] = None, title_size: Optional[int] = 18,
                        style: Optional[str] = None,
                        logscale: bool = False,
                        invert_yaxis: bool = True,
                        return_fig: bool = True, figsize: Tuple[int, int] = (6, 10),
                        fig: Optional[plt.Figure] = None, ax: Optional[plt.Axes] = None,
                        ax_pos: Tuple[int, int, int] = (1, 1, 1),
                        loglevel: str = "WARNING"):
    """
    Plots a vertical line plot.

    Args:
        data (xr.DataArray or list of xr.DataArray): DataArray(s) to plot.
        ref_data (xr.DataArray, optional): Reference DataArray to plot as dashed line.
        lev_name (str): Name of the vertical levels (default 'plev').
        lev_min (float, optional): Minimum vertical level to plot.
        lev_max (float, optional): Maximum vertical level to plot.
        labels (list of str, optional): Labels for each DataArray.
        ref_label (str, optional): reference data label.
        label_size (int, optional): Legend font size. Default is 14.
        axis_label_size (int, optional): Axis label font size. Default is 14.
        title (str, optional): Plot title.
        title_size (int, optional): Title font size. Default is 16.
        style (str, optional): Plot style (default aqua style).
        logscale (bool, optional): Use log scale for y-axis if True.
        invert_yaxis (bool, optional): Invert y-axis if True (default True).
        return_fig (bool, optional): If True, return (fig, ax).
        figsize (Tuple[int, int], optional): Figure size.
        fig (plt.Figure, optional): Optional figure to plot on.
        ax (plt.Axes, optional): Optional axes to plot on.
        ax_pos (Tuple[int, int, int], optional): Position of subplot.
        loglevel (str, optional): Logging level.
    """
    logger = log_configure(loglevel, "plot_vertical_line")
    ConfigStyle(style=style, loglevel=loglevel)

    data = to_list(data)
    labels = to_list(labels)

    # Prepare figure and axis
    fig = fig or plt.figure(figsize=figsize)
    ax = ax or fig.add_subplot(*ax_pos)

    # Select vertical levels
    if lev_min is None or lev_max is None:
        lev_min = min(data[lev_name].min().item() for data in data)
        lev_max = max(data[lev_name].max().item() for data in data)
        if ref_data is not None:
            lev_min = min(lev_min, ref_data[lev_name].min().item())
            lev_max = max(lev_max, ref_data[lev_name].max().item())
    else: # If at least one is provided, limit to the provided range
        ax.set_ylim(lev_max, lev_min)

    logger.debug(f"Plotting vertical line from {lev_min} to {lev_max} {lev_name}")

    if logscale:
        ax.set_yscale("log")

    units = data[0][lev_name].attrs.get("units", "")
    var_name = data[0].long_name or data[0].short_name
    var_units = data[0].attrs.get("units", "")

    units_latex = unit_to_latex(units) if units else ""

    if var_name and var_units:
        var_units_latex = unit_to_latex(var_units)
        xlabel = f"{var_name} ({var_units_latex})"
    else:
        xlabel = "Unknown variable"
        
    ylabel = f"{lev_name} ({units_latex})" if units_latex else lev_name  # Replace 'units' with actual units if available

    ax.set_ylabel(ylabel, fontsize=axis_label_size)
    ax.set_xlabel(xlabel, fontsize=axis_label_size)

    for i, d in enumerate(data):
        mask = (d[lev_name] >= lev_min) & (d[lev_name] <= lev_max)
        label = labels[i] if labels else None
        ax.plot(d.where(mask), d[lev_name].where(mask), label=label)

    logger.debug("Plotting reference data" if ref_data is not None else "No reference data to plot")
    
    if ref_data is not None:
        mask = (ref_data[lev_name] >= lev_min) & (ref_data[lev_name] <= lev_max)
        ax.plot(ref_data.where(mask), ref_data[lev_name].where(mask), label=ref_label, linestyle='--', color='black')

    ax.legend(fontsize=label_size)
    ax.grid(True, axis='y')

    if invert_yaxis:
        ax.invert_yaxis()

    if title:
        ax.set_title(title, fontsize=title_size)

    if return_fig:
        logger.debug("Returning figure and axes")
        return fig, ax



"""Timmean mixin for the Reader class"""
import pandas as pd
import xarray as xr
import numpy as np
from functools import partial
from aqua.core.util import check_chunk_completeness, check_seasonal_chunk_completeness, frequency_string_to_pandas
from aqua.core.util import extract_literal_and_numeric
from aqua.core.logger import log_history, log_configure
from aqua.core.histogram import histogram


class TimStat():
    """
    Time statistic AQUA module
    """


    def __init__(self, loglevel='WARNING'):
        self.loglevel = loglevel
        self.orig_freq = None
        self.logger = log_configure(loglevel, 'TimStat')

    @property
    def AVAILABLE_STATS(self):
        """Return the list of available statistics."""
        return ['mean', 'std', 'max', 'min', 'sum', 'histogram']

    def timstat(self, data, stat='mean', freq=None, exclude_incomplete=False,
                time_bounds=False, center_time=False, func_kwargs={}, **kwargs):
        """
        Compute a time statistic on the input data. The statistic is computed over a time window defined by the frequency
        parameter. The frequency can be a string (e.g. '1D', '1M', '1Y', 'QS-DEC') or a pandas frequency object. The statistic can be
        'mean', 'std', 'max', 'min' or 'histogram'. The output is a new xarray dataset with the time dimension resampled to the desired
        frequency and the statistic computed over the time window.

        Args: 
            data (xarray.Dataset): Input data to compute the statistic on.
            stat (str, func): Statistic to compute. Can be a string in ['mean', 'std', 'max', 'min', 'histogram'] or a custom function.
            freq (str): Frequency to resample the data to. Can be a string (e.g. '1D', '1M', '1Y') or a pandas frequency object.
            exclude_incomplete (bool): If True, exclude incomplete chunks from the output.
            time_bounds (bool): If True, add time bounds to the output data.
            center_time (bool): If True, center the time axis of the output data.
            func_kwargs (dict): Additional keyword arguments to pass to the custom function if stat is callable.
            kwargs (dict): Additional keyword arguments to pass to the statistic function.

        Returns:
            xarray.Dataset: Output data with the required statistic computed at the desired frequency.
        """

        if isinstance(stat, str) and stat not in self.AVAILABLE_STATS:
            raise KeyError(f'{stat} is not a statistic supported by AQUA')

        if not isinstance(stat, str) and not callable(stat):
            raise TypeError('stat must be a string or a callable function')
        
        if stat == 'histogram':  # convert to callable function
            stat = histogram
        
        resample_freq = frequency_string_to_pandas(freq)

        # disabling all options if total averaging is selected
        if resample_freq is None:
            exclude_incomplete = False
            center_time = False
            time_bnds = False

        if 'time' not in data.dims:
            raise ValueError(f'Time dimension not found in the input data. Cannot compute tim{stat} statistic')

        # Get original frequency (for history)
        if len(data.time) > 1:
            time_values = pd.to_datetime(data['time'].values[:2])
            self.orig_freq = pd.tseries.frequencies.to_offset(time_values[1] - time_values[0])
            #orig_freq = (time_values[1] - time_values[0]).total_seconds() / 3600
            #self.orig_freq = round(orig_freq)
        else:
            # this block is likely never run, as the check for time dimension is done before
            self.logger.warning('A single timestep is available, is this correct?')
            self.orig_freq = 'Unknown'
            if exclude_incomplete:
                self.logger.warning('Disabling exclude incomplete since it cannot work if we have a single tstep!')
                exclude_incomplete = False

        # if we have a frequency
        if resample_freq is not None:
            try:
                # Resample to the desired frequency
                resample_data = data.resample(time=resample_freq)
            except ValueError as exc:
                raise ValueError(f'Cant find a frequency to resample, using resample_freq={resample_freq} not work, aborting!') from exc
            
        # if frequency is undefined, meaning that we operate on the entire set
        else:
            resample_data = data

        # compact call, equivalent of "out = resample_data.mean()""
        if isinstance(stat, str):  # we already checked if it is one of the allowable stats
            self.logger.info(f'Resampling to %s frequency and computing {stat}...', str(resample_freq))
            # use the kwargs to feed the time dimension to define the method and its options
            extra_kwargs = {} if resample_freq is not None else {'dim': 'time'}
            out = getattr(resample_data, stat)(**extra_kwargs)
        else:  # we can safely assume that it is a callable function now
            self.logger.info(f'Resampling to %s frequency and computing custom function...', str(resample_freq))
            if resample_freq is not None:
                out = resample_data.map(partial(stat, **func_kwargs, **kwargs))
            else:
                out = stat(resample_data, **func_kwargs, **kwargs)

        if exclude_incomplete and freq not in [None]:
            self.logger.info('Checking if incomplete chunks has been produced...')
            if 'Q' in resample_freq:
                boolean_mask = check_seasonal_chunk_completeness(data,
                                                                 resample_frequency=resample_freq,
                                                                 loglevel=self.loglevel)
            else:
                boolean_mask = check_chunk_completeness(data,
                                                        resample_frequency=resample_freq,
                                                        loglevel=self.loglevel)
            out = out.where(boolean_mask, drop=True)

        # Set time:
        # if not center_time as the first timestamp of each month/day according to the sampling frequency
        # if center_time as the middle timestamp of each month/day according to the sampling frequency
        if center_time:
            out = self.center_time_axis(out, resample_freq)

        # Check time is correct
        if resample_freq is not None:
            if np.any(np.isnat(out.time)):
                raise ValueError('Resampling cannot produce output for all frequency step, is your input data correct?')

        out = log_history(out, f"resampled from frequency {self.orig_freq} to frequency {freq} by AQUA tim{stat}")

        # Add a variable to create time_bounds
        if time_bounds:
            resampled = data.time.resample(time=resample_freq)
            time_bnds = xr.concat([resampled.min(), resampled.max()], dim='bnds', coords='different').transpose()
            time_bnds['time'] = out.time
            time_bnds.name = 'time_bnds'
            out = xr.merge([out, time_bnds])
            if np.any(np.isnat(out.time_bnds)):
                raise ValueError('Resampling cannot produce output for all time_bnds step!')
            log_history(out, f"time_bnds added by by AQUA tim{stat}")

        return out
    
    # this is not yet a great solution, but is more general than the previous one
    def center_time_axis(self, avg_data: xr.Dataset, resample_freq: str) -> xr.Dataset:
        """
        Move the time axis of the averaged data toward the center of the averaging period.

        Args:
            avg_data (xr.Dataset): The dataset with averaged data.
            resample_freq (str): The resampling frequency (e.g., '1D', '1M', '1Y').

        Returns:
            xr.Dataset: The dataset with the time axis centered.
        """

        literal, numeric = extract_literal_and_numeric(resample_freq)
        self.logger.debug('Frequency is %s with numeric part %s', literal, numeric)

        if literal in ["M", "Y", "ME", "YE"]:
            raise ValueError(f"Centering not implemented for frequency '{resample_freq}'")

        def average_datetimeindex(idx1: pd.DatetimeIndex, idx2: pd.DatetimeIndex) -> pd.DatetimeIndex:
            return pd.to_datetime((idx1.view("int64") + idx2.view("int64")) // 2)
        
        offset = pd.tseries.frequencies.to_offset(resample_freq)

        avg_data['time'] = average_datetimeindex(pd.to_datetime(avg_data['time']),
                              pd.to_datetime(avg_data['time']) + offset)
        
        return avg_data



import cartopy.crs as ccrs

projections_maps = {
    "plate_carree": ccrs.PlateCarree,
    "albers_equal_area": ccrs.AlbersEqualArea,
    "azimuthal_equidistant": ccrs.AzimuthalEquidistant,
    "equidist_conic": ccrs.EquidistantConic,
    "lambert_conformal": ccrs.LambertConformal,
    "lambert_cylindrical": ccrs.LambertCylindrical,
    "mercator": ccrs.Mercator,
    "miller": ccrs.Miller,
    "mollweide": ccrs.Mollweide,
    "orthographic": ccrs.Orthographic,
    "robinson": ccrs.Robinson,
    "sinusoidal": ccrs.Sinusoidal,
    "stereographic": ccrs.Stereographic,
    "transverse_mercator": ccrs.TransverseMercator,
    "interrupted_gh": ccrs.InterruptedGoodeHomolosine,
    "rotated_pole": ccrs.RotatedPole,
    "europp": ccrs.EuroPP,
    "geostationary": ccrs.Geostationary,
    "nearside": ccrs.NearsidePerspective,
    "nh_polar_stereo": ccrs.NorthPolarStereo,
    "sh_polar_stereo": ccrs.SouthPolarStereo,
    "eckert_ii": ccrs.EckertII,
    "eckert_iv": ccrs.EckertIV,
    "eckert_vi": ccrs.EckertVI,
    "aitoff": ccrs.Aitoff,
    "equal_earth": ccrs.EqualEarth,
    "hammer": ccrs.Hammer,
    "lambert_equal_area": ccrs.LambertAzimuthalEqualArea,
    "gnomonic": ccrs.Gnomonic
    }

def get_projection(projname: str, **kwargs) -> ccrs.Projection:
    """
    Return a Cartopy projection by name. Refer to the Cartopy 
    documentation (https://scitools.org.uk/cartopy/) to review the 
    supported keyword arguments for each projection class.

    Args:
        projname (str): Name of the projection.
        **kwargs: Additional keyword arguments passed to the projection class.

    Returns:
        cartopy.crs.Projection: An instance of the projection.
    """
    projname = projname.lower()
    if projname not in projections_maps:
        raise ValueError(f"Unsupported projection: '{projname}'. "
                         f"Available options are: {list(projections_maps.keys())}")
    
    projection_class = projections_maps[projname]
    return projection_class(**kwargs)



"""Strategies for fixing issues in the code."""

from datetime import timedelta
import numpy as np
import pandas as pd
import xarray as xr

from aqua.core.logger import log_history, log_configure
from aqua.core.util import to_list, normalize_units

class FixerOperator:
    """
    Base class for fix operators.
    Fix operators are method which apply specific fixes to the data.
    They all operate on xarray DataArray or Dataset objects.
    The FixOperator class is initialized with a dictionary of fixes and a log level.

    Args:
        fixes (dict): Dictionary containing the fixes to be applied.
        loglevel (int, optional): Log level for logging. Defaults to None.
    """

    def __init__(self, fixes, loglevel=None):

        self.fixes = fixes
        self.loglevel = loglevel
        self.logger = log_configure(log_level = loglevel, log_name="FixerOperator")

    def apply_unit_fix(self, data, time_correction=False):
        """
        Applies unit fixes stored in variable attributes (target_units, factor and offset)

        Arguments:
            data (xr.DataArray):  input DataArray
        """
        tgt_units = data.attrs.get("tgt_units", None)
        org_units = data.attrs.get("units", None)
        self.logger.debug("%s: org_units is %s, tgt_units is %s",
                        data.name, org_units, tgt_units)

        # if units are not already updated and if a tgt_units exist
        if tgt_units and org_units != tgt_units:
            self.logger.debug("Applying unit fixes for %s ", data.name)

            # define an old units
            data.attrs.update({"src_units": org_units, "units_fixed": 1})
            data.attrs["units"] = normalize_units(tgt_units)
            factor = data.attrs.get("factor", 1)
            offset = data.attrs.get("offset", 0)
            time_conversion_flag = data.attrs.get("time_conversion_flag", 0)
            if factor != 1:
                data *= factor
                # if a special dpm correction has been defined, apply it
                if time_conversion_flag and time_correction is not False:
                    data /=  time_correction
            if offset != 0:
                data += offset
            log_history(data, f"Units changed to {tgt_units} by fixer")
            data.attrs.pop('tgt_units', None)

    def delete_variables(self, data):
        """
        Remove variables which are set to be deleted in the fixer
        """

        # remove variables which should be deleted
        dellist = [x for x in to_list(self.fixes.get("delete", [])) if x in data.variables]
        if dellist:
            data = data.drop_vars(dellist)

        return data

    def timeshifter(self, data):
        """
        Apply a timeshift to the time coordinate of an xr.Dataset.

        Parameters:
        - data (xr.Dataset): The dataset containing a 'time' coordinate to be shifted.

        Returns:
        - xr.Dataset: The dataset with the 'time' coordinate shifted based on the specified timeshift
                      which is retrieved from the fixes dictionary.
        """
        timeshift = self.fixes.get('timeshift', None)

        if timeshift is None:
            return data

        if 'time' not in data:
            raise KeyError("'time' coordinate not found in the dataset.")

        field = data.copy()
        if isinstance(timeshift, int):
            self.logger.info('Shifting the time axis by %s timesteps.', timeshift)
            time_interval = timeshift * data.time.diff("time").isel(time=0).values
            field = field.assign_coords(time=data.time + time_interval)
        elif isinstance(timeshift, str):
            self.logger.info('Shifting time axis by %s following pandas timedelta.', timeshift)
            field['time'] = field['time'] + pd.Timedelta(timeshift)
        else:
            raise TypeError('timeshift should be either a integer (timesteps) or a pandas Timedelta!')

        return field

    def wrapper_decumulate(self, data, deltat, variables, varlist, jump):
        """
        Wrapper function for decumulation, which takes into account the requirement of
        keeping into memory the last step for streaming/fdb purposes

        Args:
            Data: Xarray Dataset
            variables: The fixes of the variables
            varlist: the variable dictionary with the old and new names
            jump: the jump for decumulation

        Returns:
            Dataset with decumulated fixes
        """

        for var in variables:
            # Decumulate if required
            if variables[var].get("decumulate", None):
                varname = varlist[var]
                if varname in data.variables:
                    self.logger.debug("Starting decumulation for variable %s", varname)
                    keep_first = variables[var].get("keep_first", True)
                    data[varname] = self.simple_decumulate(data[varname],
                                                           deltat=deltat,
                                                           jump=jump,
                                                           keep_first=keep_first)
                    log_history(data[varname], f"Variable {varname} decumulated by fixer")

        return data
    
    def simple_decumulate(self, data, deltat=3600, jump=None, keep_first=True):
        """
        Remove cumulative effect on IFS fluxes.

        Args:
            data (xr.DataArray):     field to be processed
            jump (str):              used to fix periodic jumps (a very specific NextGEMS IFS issue)
                                    Examples: jump='month' (the NextGEMS case), jump='day')
            keep_first (bool):       if to keep the first value as it is (True) or place a 0 (False)

        Returns:
            A xarray.DataArray where the cumulation has been removed
        """

        # get the derivatives
        deltas = data.diff(dim='time')

        # add a first timestep empty to align the original and derived fields

        if keep_first:
            zeros = data.isel(time=0)
        else:
            zeros = xr.zeros_like(data.isel(time=0))

        deltas = xr.concat([zeros, deltas], dim='time', coords='different', compat='equals').transpose('time', ...)

        if jump:
            # universal mask based on the change of month (shifted by one timestep)
            dt = np.timedelta64(timedelta(seconds=deltat))
            data1 = data.assign_coords(time=data.time - dt)
            data2 = data.assign_coords(time=data1.time - dt)
            # Mask of dates where month changed in the previous timestep
            mask = data1[f'time.{jump}'].assign_coords(time=data.time) == data2[f'time.{jump}'].assign_coords(time=data.time)

            # kaboom: exploit where
            deltas = deltas.where(mask, data)

        # add an attribute that can be later used to infer about decumulation
        deltas.attrs['decumulated'] = 1

        return deltas

    def wrapper_nanfirst(self, data, variables, varlist, startdate=None, enddate=None):
        """
        Wrapper function for settting to nan first step of each month.
        This allows to fix an issue with IFS data where the first step of each month is corrupted.

        Args:
            Data: Xarray Dataset
            variables: The fixes of the variables
            varlist: the variable dictionary with the old and new names
            startdate: date before which to fix the first timestep of each month (could be False)
            enddate: date after which to fix the first timestep of each month (could be False)

        Returns:
            Dataset with data on first step of each month set to NaN
        """

        for var in variables:
            fix = variables[var].get("nanfirst", False)
            if fix:
                varname = varlist[var]
                if varname in data.variables:
                    self.logger.debug("Setting first step of months before %s and after %s to NaN for variable %s",
                                      enddate, startdate, varname)
                    log_history(data[varname], f"Fixer set first step of months before {enddate} and after {startdate} to NaN")
                    data[varname] = self.nanfirst(data[varname], startdate=startdate, enddate=enddate)

        return data

    def nanfirst(self, data, startdate=False, enddate=False):
        """
        Set to NaN the first step of each month before and/or after a given date for an xarray

        Args:
            data: Xarray DataArray
            startdate: date before which to fix the first timestep of each month (defaults to False)
            enddate: date after which to fix the first timestep of each month (defaults to False)

        Returns:
            DataArray in with data on first step of each month is set to NaN
        """

        first = data.time.groupby(data['time.year']*100+data['time.month']).first()
        if enddate:
            first = first.where(first < np.datetime64(str(enddate)), drop=True)
        if startdate:
            first = first.where(first > np.datetime64(str(startdate)), drop=True)
        mask = data.time.isin(first)
        data = data.where(~mask, np.nan)

        return data